from textual.app import App, ComposeResult
from textual.suggester import SuggestFromList
from textual.widgets import Input


fruits = ["apple", "pear", "mango", "peach", "strawberry", "blueberry", "banana"]


class FruitsApp(App[None]):
    CSS = """
    Input > .input--suggestion {
        color: red;
        text-style: italic;
    }
    """

    def compose(self) -> ComposeResult:
        yield Input("straw", suggester=SuggestFromList(fruits))
        yield Input("straw", suggester=SuggestFromList(fruits))
        yield Input("p", suggester=SuggestFromList(fruits))
        yield Input("b", suggester=SuggestFromList(fruits))
        yield Input("a", suggester=SuggestFromList(fruits))


if __name__ == "__main__":
    app = FruitsApp()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import Container
from textual.widgets import Label


class AlignAllApp(App):
    """App that illustrates all alignments."""

    CSS_PATH = "align_all.tcss"

    def compose(self) -> ComposeResult:
        yield Container(Label("left top"), id="left-top")
        yield Container(Label("center top"), id="center-top")
        yield Container(Label("right top"), id="right-top")
        yield Container(Label("left middle"), id="left-middle")
        yield Container(Label("center middle"), id="center-middle")
        yield Container(Label("right middle"), id="right-middle")
        yield Container(Label("left bottom"), id="left-bottom")
        yield Container(Label("center bottom"), id="center-bottom")
        yield Container(Label("right bottom"), id="right-bottom")


if __name__ == "__main__":
    AlignAllApp().run()



import pytest

from textual.app import App, ComposeResult
from textual.geometry import Offset
from textual.widgets import TextArea
from textual.widgets.text_area import Selection

TEXT = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
"""


class TextAreaApp(App):
    def __init__(self, read_only: bool = False):
        super().__init__()
        self.read_only = read_only

    def compose(self) -> ComposeResult:
        yield TextArea(TEXT, show_line_numbers=True, read_only=self.read_only)


@pytest.fixture(params=[True, False])
async def app(request):
    """Each test that receives an `app` will execute twice.
    Once with read_only=True, and once with read_only=False.
    """
    return TextAreaApp(read_only=request.param)


async def test_mouse_click(app: TextAreaApp):
    """When you click the TextArea, the cursor moves to the expected location."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        await pilot.click(TextArea, Offset(x=5, y=2))
        assert text_area.selection == Selection.cursor((1, 0))


async def test_mouse_click_clamp_from_right(app: TextAreaApp):
    """When you click to the right of the document bounds, the cursor is clamped
    to within the document bounds."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        await pilot.click(TextArea, Offset(x=8, y=20))
        assert text_area.selection == Selection.cursor((4, 0))


async def test_mouse_click_gutter_clamp(app: TextAreaApp):
    """When you click the gutter, it selects the start of the line."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        await pilot.click(TextArea, Offset(x=0, y=3))
        assert text_area.selection == Selection.cursor((2, 0))


async def test_cursor_movement_basic():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("01234567\n012345\n0123456789")

        await pilot.press("right")
        assert text_area.selection == Selection.cursor((0, 1))

        await pilot.press("down")
        assert text_area.selection == Selection.cursor((1, 1))

        await pilot.press("left")
        assert text_area.selection == Selection.cursor((1, 0))

        await pilot.press("up")
        assert text_area.selection == Selection.cursor((0, 0))


async def test_cursor_selection_right(app: TextAreaApp):
    """When you press shift+right the selection is updated correctly."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        await pilot.press(*["shift+right"] * 3)
        assert text_area.selection == Selection((0, 0), (0, 3))


async def test_cursor_selection_right_to_previous_line(app: TextAreaApp):
    """When you press shift+right resulting in the cursor moving to the next line,
    the selection is updated correctly."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.selection = Selection.cursor((0, 15))
        await pilot.press(*["shift+right"] * 4)
        assert text_area.selection == Selection((0, 15), (1, 2))


async def test_cursor_selection_left(app: TextAreaApp):
    """When you press shift+left the selection is updated correctly."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.selection = Selection.cursor((2, 5))
        await pilot.press(*["shift+left"] * 3)
        assert text_area.selection == Selection((2, 5), (2, 2))


async def test_cursor_selection_left_to_previous_line(app: TextAreaApp):
    """When you press shift+left resulting in the cursor moving back to the previous line,
    the selection is updated correctly."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.selection = Selection.cursor((2, 2))
        await pilot.press(*["shift+left"] * 3)

        # The cursor jumps up to the end of the line above.
        end_of_previous_line = len(TEXT.splitlines()[1])
        assert text_area.selection == Selection((2, 2), (1, end_of_previous_line))


async def test_cursor_selection_up(app: TextAreaApp):
    """When you press shift+up the selection is updated correctly."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.move_cursor((2, 3))

        await pilot.press("shift+up")
        assert text_area.selection == Selection((2, 3), (1, 3))


async def test_cursor_selection_up_when_cursor_on_first_line(app: TextAreaApp):
    """When you press shift+up the on the first line, it selects to the start."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.move_cursor((0, 4))

        await pilot.press("shift+up")
        assert text_area.selection == Selection((0, 4), (0, 0))
        await pilot.press("shift+up")
        assert text_area.selection == Selection((0, 4), (0, 0))


async def test_cursor_selection_down(app: TextAreaApp):
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.move_cursor((2, 5))

        await pilot.press("shift+down")
        assert text_area.selection == Selection((2, 5), (3, 5))


async def test_cursor_selection_down_when_cursor_on_last_line(app: TextAreaApp):
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("ABCDEF\nGHIJK")
        text_area.move_cursor((1, 2))

        await pilot.press("shift+down")
        assert text_area.selection == Selection((1, 2), (1, 5))
        await pilot.press("shift+down")
        assert text_area.selection == Selection((1, 2), (1, 5))


async def test_cursor_word_right(app: TextAreaApp):
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("ABC DEF\nGHIJK")

        await pilot.press("ctrl+right")

        assert text_area.selection == Selection.cursor((0, 3))


async def test_cursor_word_right_select(app: TextAreaApp):
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("ABC DEF\nGHIJK")

        await pilot.press("ctrl+shift+right")

        assert text_area.selection == Selection((0, 0), (0, 3))


async def test_cursor_word_left(app: TextAreaApp):
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("ABC DEF\nGHIJK")
        text_area.move_cursor((0, 7))

        await pilot.press("ctrl+left")

        assert text_area.selection == Selection.cursor((0, 4))


async def test_cursor_word_left_select(app: TextAreaApp):
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("ABC DEF\nGHIJK")
        text_area.move_cursor((0, 7))

        await pilot.press("ctrl+shift+left")

        assert text_area.selection == Selection((0, 7), (0, 4))


@pytest.mark.parametrize("key", ["end", "ctrl+e"])
async def test_cursor_to_line_end(key, app: TextAreaApp):
    """You can use the keyboard to jump the cursor to the end of the current line."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.selection = Selection.cursor((2, 2))
        await pilot.press(key)
        eol_index = len(TEXT.splitlines()[2])
        assert text_area.cursor_location == (2, eol_index)
        assert text_area.selection.is_empty


@pytest.mark.parametrize("key", ["home", "ctrl+a"])
async def test_cursor_to_line_home_basic_behaviour(key, app: TextAreaApp):
    """You can use the keyboard to jump the cursor to the start of the current line."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.selection = Selection.cursor((2, 2))
        await pilot.press(key)
        assert text_area.cursor_location == (2, 0)
        assert text_area.selection.is_empty


@pytest.mark.parametrize(
    "cursor_start,cursor_destination",
    [
        ((0, 0), (0, 4)),
        ((0, 2), (0, 0)),
        ((0, 4), (0, 0)),
        ((0, 5), (0, 4)),
        ((0, 9), (0, 4)),
        ((0, 15), (0, 4)),
    ],
)
async def test_cursor_line_home_smart_home(
    cursor_start, cursor_destination, app: TextAreaApp
):
    """If the line begins with whitespace, pressing home firstly goes
    to the start of the (non-whitespace) content. Pressing it again takes you to column
    0. If you press it again, it goes back to the first non-whitespace column."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("    hello world")
        text_area.move_cursor(cursor_start)
        await pilot.press("home")
        assert text_area.selection == Selection.cursor(cursor_destination)


async def test_cursor_page_down(app: TextAreaApp):
    """Pagedown moves the cursor down 1 page, retaining column index."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("XXX\n" * 200)
        text_area.selection = Selection.cursor((0, 1))
        await pilot.press("pagedown")
        margin = 2
        assert text_area.selection == Selection.cursor(
            (app.console.height - 1 - margin, 1)
        )


async def test_cursor_page_up(app: TextAreaApp):
    """Pageup moves the cursor up 1 page, retaining column index."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("XXX\n" * 200)
        text_area.selection = Selection.cursor((100, 1))
        await pilot.press("pageup")
        margin = 2
        assert text_area.selection == Selection.cursor(
            (100 - app.console.height + 1 + margin, 1)
        )


async def test_cursor_vertical_movement_visual_alignment_snapping(app: TextAreaApp):
    """When you move the cursor vertically, it should stay vertically
    aligned even when double-width characters are used."""
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.text = "こんにちは\n012345"
        text_area.move_cursor((1, 3), record_width=True)

        # The '3' is aligned with ん at (0, 1)
        # こんにちは
        # 012345
        # Pressing `up` takes us from (1, 3) to (0, 1) because record_width=True.
        await pilot.press("up")
        assert text_area.selection == Selection.cursor((0, 1))

        # Pressing `down` takes us from (0, 1) to (1, 3)
        await pilot.press("down")
        assert text_area.selection == Selection.cursor((1, 3))


async def test_select_line_binding(app: TextAreaApp):
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.move_cursor((2, 2))

        await pilot.press("f6")

        assert text_area.selection == Selection((2, 0), (2, 56))


async def test_select_all_binding(app: TextAreaApp):
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)

        await pilot.press("f7")

        assert text_area.selection == Selection((0, 0), (4, 0))



from __future__ import annotations

from typing import Callable, TypeVar

from textual.css.model import SelectorSet
from textual.css.parse import parse_selectors
from textual.css.tokenizer import TokenError
from textual.message import Message

DecoratedType = TypeVar("DecoratedType")


class OnDecoratorError(Exception):
    """Errors related to the `on` decorator.

    Typically raised at import time as an early warning system.
    """


class OnNoWidget(Exception):
    """A selector was applied to an attribute that isn't a widget."""


def on(
    message_type: type[Message], selector: str | None = None, **kwargs: str
) -> Callable[[DecoratedType], DecoratedType]:
    """Decorator to declare that the method is a message handler.

    The decorator accepts an optional CSS selector that will be matched against a widget exposed by
    a `control` property on the message.

    Example:
        ```python
        # Handle the press of buttons with ID "#quit".
        @on(Button.Pressed, "#quit")
        def quit_button(self) -> None:
            self.app.quit()
        ```

    Keyword arguments can be used to match additional selectors for attributes
    listed in [`ALLOW_SELECTOR_MATCH`][textual.message.Message.ALLOW_SELECTOR_MATCH].

    Example:
        ```python
        # Handle the activation of the tab "#home" within the `TabbedContent` "#tabs".
        @on(TabbedContent.TabActivated, "#tabs", pane="#home")
        def switch_to_home(self) -> None:
            self.log("Switching back to the home tab.")
            ...
        ```

    Args:
        message_type: The message type (i.e. the class).
        selector: An optional [selector](/guide/CSS#selectors). If supplied, the handler will only be called if `selector`
            matches the widget from the `control` attribute of the message.
        **kwargs: Additional selectors for other attributes of the message.
    """

    selectors: dict[str, str] = {}
    if selector is not None:
        selectors["control"] = selector
    if kwargs:
        selectors.update(kwargs)

    parsed_selectors: dict[str, tuple[SelectorSet, ...]] = {}
    for attribute, css_selector in selectors.items():
        if attribute == "control":
            if message_type.control == Message.control:
                raise OnDecoratorError(
                    "The message class must have a 'control' to match with the on decorator"
                )
        elif attribute not in message_type.ALLOW_SELECTOR_MATCH:
            raise OnDecoratorError(
                f"The attribute {attribute!r} can't be matched; have you added it to "
                + f"{message_type.__name__}.ALLOW_SELECTOR_MATCH?"
            )
        try:
            parsed_selectors[attribute] = parse_selectors(css_selector)
        except TokenError:
            raise OnDecoratorError(
                f"Unable to parse selector {css_selector!r} for {attribute}; check for syntax errors"
            ) from None

    def decorator(method: DecoratedType) -> DecoratedType:
        """Store message and selector in function attribute, return callable unaltered."""

        if not hasattr(method, "_textual_on"):
            setattr(method, "_textual_on", [])
        getattr(method, "_textual_on").append((message_type, parsed_selectors))

        return method

    return decorator



from textual.app import App, ComposeResult
from textual.widgets import Label


class BorderTitleApp(App):
    CSS_PATH = "border_title_colors.tcss"

    def compose(self) -> ComposeResult:
        yield Label("Hello, World!")

    def on_mount(self) -> None:
        label = self.query_one(Label)
        label.border_title = "Textual Rocks"
        label.border_subtitle = "Textual Rocks"


if __name__ == "__main__":
    app = BorderTitleApp()
    app.run()



from __future__ import annotations

from typing import TYPE_CHECKING, cast

from typing_extensions import Literal

if TYPE_CHECKING:
    from typing_extensions import Final

FEATURES: Final = {"devtools", "debug", "headless"}

FeatureFlag = Literal["devtools", "debug", "headless"]


def parse_features(features: str) -> frozenset[FeatureFlag]:
    """Parse features env var

    Args:
        features: Comma separated feature flags

    Returns:
        A frozen set of known features.
    """

    features_set = frozenset(
        feature.strip().lower() for feature in features.split(",") if feature.strip()
    ).intersection(FEATURES)

    return cast("frozenset[FeatureFlag]", features_set)



from textual.app import App, ComposeResult
from textual.widgets import Digits


class DigitApp(App):
    CSS = """
    #digits1 {
        text-align: left;
    }
    #digits2 {
        text-align:center;
    }
    #digits3 {
        text-align:right;
    }
    """

    def compose(self) -> ComposeResult:
        yield Digits("3.1427", id="digits1")
        yield Digits(" 0123456789+-.,", id="digits2")
        yield Digits("3x10^4", id="digits3")


if __name__ == "__main__":
    app = DigitApp()
    app.run()



from pathlib import Path

from textual.css.stylesheet import Stylesheet


def test_mega_stylesheet() -> None:
    """It should be possible to load a known-good stylesheet."""
    mega_stylesheet = Stylesheet()
    mega_stylesheet.read(Path(__file__).parent / "test_mega_stylesheet.tcss")
    mega_stylesheet.parse()
    assert ".---we-made-it-to-the-end---" in mega_stylesheet.css



"""Test https://github.com/Textualize/textual/issues/2557"""

from textual.app import App, ComposeResult
from textual.widgets import Select, Button


class SelectRebuildApp(App[None]):

    def compose(self) -> ComposeResult:
        yield Select[int]((("1", 1), ("2", 2)))
        yield Button("Rebuild")

    def on_button_pressed(self):
        self.query_one(Select).set_options((
            ("This", 0), ("Should", 1), ("Be", 2),
            ("What", 3), ("Goes", 4), ("Into",5),
            ("The", 6), ("Snapshit", 7)
        ))

if __name__ == "__main__":
    SelectRebuildApp().run()



from textual.app import App
from textual.containers import VerticalScroll
from textual.widgets import Placeholder


class MaxWidthApp(App):
    CSS_PATH = "max_width.tcss"

    def compose(self):
        yield VerticalScroll(
            Placeholder("max-width: 50h", id="p1"),
            Placeholder("max-width: 999", id="p2"),
            Placeholder("max-width: 50%", id="p3"),
            Placeholder("max-width: 30", id="p4"),
        )


if __name__ == "__main__":
    app = MaxWidthApp()
    app.run()



from textual.app import App, ComposeResult
from textual.scroll_view import ScrollView
from textual.strip import Strip
from rich.segment import Segment


class LineWidget(ScrollView):
    COMPONENT_CLASSES = {"faded"}

    DEFAULT_CSS = """
    LineWidget {
        background:blue;
        color:white;
        &>.faded{
            text-opacity:0.5;
          
        }
    }
    """

    def render_line(self, y: int) -> Strip:
        style = self.get_component_rich_style("faded")
        return Strip([Segment("W" * self.size.width, style)])


class TestApp(App):
    def compose(self) -> ComposeResult:
        yield LineWidget()


if __name__ == "__main__":
    app = TestApp()
    app.run()



from textual.app import App, ComposeResult
from textual.widgets import Label


class SingleLabelApp(App[None]):
    """An app with a single label that's 20 x 10."""

    def compose(self) -> ComposeResult:
        yield Label(("12345678901234567890\n" * 10).strip())


if __name__ == "__main__":
    SingleLabelApp().run()



from textual.app import App, ComposeResult
from textual.containers import Horizontal, VerticalScroll
from textual.widgets import Label, RadioButton, RadioSet


class RadioSetChangedApp(App[None]):
    CSS_PATH = "radio_set_changed.tcss"

    def compose(self) -> ComposeResult:
        with VerticalScroll():
            with Horizontal():
                with RadioSet(id="focus_me"):
                    yield RadioButton("Battlestar Galactica")
                    yield RadioButton("Dune 1984")
                    yield RadioButton("Dune 2021")
                    yield RadioButton("Serenity", value=True)
                    yield RadioButton("Star Trek: The Motion Picture")
                    yield RadioButton("Star Wars: A New Hope")
                    yield RadioButton("The Last Starfighter")
                    yield RadioButton(
                        "Total Recall :backhand_index_pointing_right: :red_circle:"
                    )
                    yield RadioButton("Wing Commander")
            with Horizontal():
                yield Label(id="pressed")
            with Horizontal():
                yield Label(id="index")

    def on_mount(self) -> None:
        self.query_one(RadioSet).focus()

    def on_radio_set_changed(self, event: RadioSet.Changed) -> None:
        self.query_one("#pressed", Label).update(
            f"Pressed button label: {event.pressed.label}"
        )
        self.query_one("#index", Label).update(
            f"Pressed button index: {event.radio_set.pressed_index}"
        )


if __name__ == "__main__":
    RadioSetChangedApp().run()



from __future__ import annotations

import pytest

from textual.app import App, ComposeResult
from textual.geometry import Offset
from textual.widgets import Input

# A string containing only single-width characters
TEXT_SINGLE = "That gum you like is going to come back in style"

# A string containing only double-width characters
TEXT_DOUBLE = "こんにちは"

# A string containing both single and double-width characters
TEXT_MIXED = "aこんbcにdちeは"


class InputApp(App[None]):
    def __init__(self, text):
        super().__init__()
        self._text = text

    def compose(self) -> ComposeResult:
        yield Input(self._text)


@pytest.mark.parametrize(
    "text, click_at, should_land",
    (
        # Single-width characters
        (TEXT_SINGLE, 0, 0),
        (TEXT_SINGLE, 1, 1),
        (TEXT_SINGLE, 10, 10),
        (TEXT_SINGLE, len(TEXT_SINGLE) - 1, len(TEXT_SINGLE) - 1),
        (TEXT_SINGLE, len(TEXT_SINGLE), len(TEXT_SINGLE)),
        (TEXT_SINGLE, len(TEXT_SINGLE) + 10, len(TEXT_SINGLE)),
        # Double-width characters
        (TEXT_DOUBLE, 0, 0),
        (TEXT_DOUBLE, 1, 0),
        (TEXT_DOUBLE, 2, 1),
        (TEXT_DOUBLE, 3, 1),
        (TEXT_DOUBLE, 4, 2),
        (TEXT_DOUBLE, 5, 2),
        (TEXT_DOUBLE, (len(TEXT_DOUBLE) * 2) - 1, len(TEXT_DOUBLE) - 1),
        (TEXT_DOUBLE, len(TEXT_DOUBLE) * 2, len(TEXT_DOUBLE)),
        (TEXT_DOUBLE, len(TEXT_DOUBLE) * 10, len(TEXT_DOUBLE)),
        # Mixed-width characters
        (TEXT_MIXED, 0, 0),
        (TEXT_MIXED, 1, 1),
        (TEXT_MIXED, 2, 1),
        (TEXT_MIXED, 3, 2),
        (TEXT_MIXED, 4, 2),
        (TEXT_MIXED, 5, 3),
        (TEXT_MIXED, 13, 9),
        (TEXT_MIXED, 14, 9),
        (TEXT_MIXED, 15, 10),
        (TEXT_MIXED, 60, 10),
    ),
)
async def test_mouse_clicks_within(text, click_at, should_land):
    """Mouse clicks should result in the cursor going to the right place."""
    async with InputApp(text).run_test() as pilot:
        # Note the offsets to take into account the decoration around an
        # Input.
        await pilot.click(Input, Offset(click_at + 3, 1))
        await pilot.pause()
        assert pilot.app.query_one(Input).cursor_position == should_land


async def test_mouse_click_outwith():
    """Mouse clicks outside the input should not affect cursor position."""
    async with InputApp(TEXT_SINGLE).run_test() as pilot:
        pilot.app.query_one(Input).cursor_position = 3
        assert pilot.app.query_one(Input).cursor_position == 3
        await pilot.click(Input, Offset(0, 0))
        await pilot.pause()
        assert pilot.app.query_one(Input).cursor_position == 3



"""Provides classes for holding and managing notifications."""

from __future__ import annotations

from dataclasses import dataclass, field
from time import time
from typing import Iterator
from uuid import uuid4

from rich.repr import Result
from typing_extensions import Literal, Self, TypeAlias

from textual.message import Message

SeverityLevel: TypeAlias = Literal["information", "warning", "error"]
"""The severity level for a notification."""


@dataclass
class Notify(Message, bubble=False):
    """Message to show a notification."""

    notification: Notification


@dataclass
class Notification:
    """Holds the details of a notification."""

    message: str
    """The message for the notification."""

    title: str = ""
    """The title for the notification."""

    severity: SeverityLevel = "information"
    """The severity level for the notification."""

    timeout: float = 5
    """The timeout (in seconds) for the notification."""

    raised_at: float = field(default_factory=time)
    """The time when the notification was raised (in Unix time)."""

    identity: str = field(default_factory=lambda: str(uuid4()))
    """The unique identity of the notification."""

    @property
    def time_left(self) -> float:
        """The time left until this notification expires"""
        return (self.raised_at + self.timeout) - time()

    @property
    def has_expired(self) -> bool:
        """Has the notification expired?"""
        return self.time_left <= 0

    def __rich_repr__(self) -> Result:
        yield "message", self.message
        yield "title", self.title, ""
        yield "severity", self.severity
        yield "raised_it", self.raised_at
        yield "identity", self.identity
        yield "time_left", self.time_left
        yield "has_expired", self.has_expired


class Notifications:
    """Class for managing a collection of notifications."""

    def __init__(self) -> None:
        """Initialise the notification collection."""
        self._notifications: dict[str, Notification] = {}

    def _reap(self) -> Self:
        """Remove any expired notifications from the notification collection."""
        for notification in list(self._notifications.values()):
            if notification.has_expired:
                del self._notifications[notification.identity]
        return self

    def add(self, notification: Notification) -> Self:
        """Add the given notification to the collection of managed notifications.

        Args:
            notification: The notification to add.

        Returns:
            Self.
        """
        self._reap()._notifications[notification.identity] = notification
        return self

    def clear(self) -> Self:
        """Clear all the notifications."""
        self._notifications.clear()
        return self

    def __len__(self) -> int:
        """The number of notifications."""
        return len(self._reap()._notifications)

    def __iter__(self) -> Iterator[Notification]:
        return iter(self._reap()._notifications.values())

    def __contains__(self, notification: Notification) -> bool:
        return notification.identity in self._notifications

    def __delitem__(self, notification: Notification) -> None:
        try:
            del self._reap()._notifications[notification.identity]
        except KeyError:
            # An attempt to remove a notification we don't know about is a
            # no-op. What matters here is that the notification is forgotten
            # about, and it looks like a caller has tried to be
            # belt-and-braces. We're fine with this.
            pass



from textual.app import App, ComposeResult
from textual.containers import Vertical
from textual.widgets import Label


class Width100PCentApp(App[None]):

    CSS = """
    Vertical {
        border: solid red;
        width: auto;

        Label {
            border: solid green;
        }

        #first {
            width: 100%;
        }

        #second {
            width: auto;
        }
    }
    """

    def compose(self) -> ComposeResult:
        with Vertical():
            yield Label("I want to be 100% of my parent", id="first")
            yield Label(
                "I want my parent to be wide enough to wrap me and no more", id="second"
            )


if __name__ == "__main__":
    Width100PCentApp().run()



"""
Tools for lazy loading widgets.
"""

from __future__ import annotations

from textual.widget import Widget


class Lazy(Widget):
    """Wraps a widget so that it is mounted *lazily*.

    Lazy widgets are mounted after the first refresh. This can be used to display some parts of
    the UI very quickly, followed by the lazy widgets. Technically, this won't make anything
    faster, but it reduces the time the user sees a blank screen and will make apps feel
    more responsive.

    Making a widget lazy is beneficial for widgets which start out invisible, such as tab panes.

    Note that since lazy widgets aren't mounted immediately (by definition), they will not appear
    in queries for a brief interval until they are mounted. Your code should take this in to account.

    Example:
        ```python
        def compose(self) -> ComposeResult:
            yield Footer()
            with ColorTabs("Theme Colors", "Named Colors"):
                yield Content(ThemeColorButtons(), ThemeColorsView(), id="theme")
                yield Lazy(NamedColorsView())
        ```

    """

    DEFAULT_CSS = """
    Lazy {
        display: none;        
    } 
    """

    def __init__(self, widget: Widget) -> None:
        """Create a lazy widget.

        Args:
            widget: A widget that should be mounted after a refresh.
        """
        self._replace_widget = widget
        super().__init__()

    def compose_add_child(self, widget: Widget) -> None:
        self._replace_widget.compose_add_child(widget)

    async def mount_composed_widgets(self, widgets: list[Widget]) -> None:
        parent = self.parent
        if parent is None:
            return
        assert isinstance(parent, Widget)

        async def mount() -> None:
            """Perform the mount and discard the lazy widget."""
            await parent.mount(self._replace_widget, after=self)
            await self.remove()

        self.call_after_refresh(mount)



"""Unit tests for the Markdown widget."""

from __future__ import annotations

from pathlib import Path
from typing import Iterator

import pytest
from markdown_it.token import Token
from rich.style import Style
from rich.text import Span

import textual.widgets._markdown as MD
from textual import on
from textual.app import App, ComposeResult
from textual.widget import Widget
from textual.widgets import Markdown
from textual.widgets.markdown import MarkdownBlock


class UnhandledToken(MarkdownBlock):
    def __init__(self, markdown: Markdown, token: Token) -> None:
        super().__init__(markdown)
        self._token = token

    def __repr___(self) -> str:
        return self._token.type


class FussyMarkdown(Markdown):
    def unhandled_token(self, token: Token) -> MarkdownBlock | None:
        return UnhandledToken(self, token)


class MarkdownApp(App[None]):
    def __init__(self, markdown: str) -> None:
        super().__init__()
        self._markdown = markdown

    def compose(self) -> ComposeResult:
        yield FussyMarkdown(self._markdown)


@pytest.mark.parametrize(
    ["document", "expected_nodes"],
    [
        # Basic markup.
        ("", []),
        ("# Hello", [MD.MarkdownH1]),
        ("## Hello", [MD.MarkdownH2]),
        ("### Hello", [MD.MarkdownH3]),
        ("#### Hello", [MD.MarkdownH4]),
        ("##### Hello", [MD.MarkdownH5]),
        ("###### Hello", [MD.MarkdownH6]),
        ("---", [MD.MarkdownHorizontalRule]),
        ("Hello", [MD.MarkdownParagraph]),
        ("Hello\nWorld", [MD.MarkdownParagraph]),
        ("> Hello", [MD.MarkdownBlockQuote, MD.MarkdownParagraph]),
        ("- One\n-Two", [MD.MarkdownBulletList, MD.MarkdownParagraph]),
        (
            "1. One\n2. Two",
            [MD.MarkdownOrderedList, MD.MarkdownParagraph, MD.MarkdownParagraph],
        ),
        ("    1", [MD.MarkdownFence]),
        ("```\n1\n```", [MD.MarkdownFence]),
        ("```python\n1\n```", [MD.MarkdownFence]),
        ("""| One | Two |\n| :- | :- |\n| 1 | 2 |""", [MD.MarkdownTable]),
        # Test for https://github.com/Textualize/textual/issues/2676
        (
            "- One\n```\nTwo\n```\n- Three\n",
            [
                MD.MarkdownBulletList,
                MD.MarkdownParagraph,
                MD.MarkdownFence,
                MD.MarkdownBulletList,
                MD.MarkdownParagraph,
            ],
        ),
    ],
)
async def test_markdown_nodes(
    document: str, expected_nodes: list[Widget | list[Widget]]
) -> None:
    """A Markdown document should parse into the expected Textual node list."""

    def markdown_nodes(root: Widget) -> Iterator[MarkdownBlock]:
        for node in root.children:
            if isinstance(node, MarkdownBlock):
                yield node
            yield from markdown_nodes(node)

    async with MarkdownApp(document).run_test() as pilot:
        assert [
            node.__class__ for node in markdown_nodes(pilot.app.query_one(Markdown))
        ] == expected_nodes


async def test_softbreak_split_links_rendered_correctly() -> None:
    """Test for https://github.com/Textualize/textual/issues/2805"""

    document = """\
My site [has
this
URL](https://example.com)\
"""
    async with MarkdownApp(document).run_test() as pilot:
        markdown = pilot.app.query_one(Markdown)
        paragraph = markdown.children[0]
        assert isinstance(paragraph, MD.MarkdownParagraph)
        assert paragraph._text.plain == "My site has this URL"
        expected_spans = [
            Span(8, 11, Style(meta={"@click": "link('https://example.com')"})),
            Span(11, 12, Style(meta={"@click": "link('https://example.com')"})),
            Span(12, 16, Style(meta={"@click": "link('https://example.com')"})),
            Span(16, 17, Style(meta={"@click": "link('https://example.com')"})),
            Span(17, 20, Style(meta={"@click": "link('https://example.com')"})),
        ]

    assert paragraph._text.spans == expected_spans


async def test_load_non_existing_file() -> None:
    """Loading a file that doesn't exist should result in the obvious error."""
    async with MarkdownApp("").run_test() as pilot:
        with pytest.raises(FileNotFoundError):
            await pilot.app.query_one(Markdown).load(
                Path("---this-does-not-exist---.it.is.not.a.md")
            )


@pytest.mark.parametrize(
    ("anchor", "found"),
    [
        ("hello-world", False),
        ("hello-there", True),
    ],
)
async def test_goto_anchor(anchor: str, found: bool) -> None:
    """Going to anchors should return a boolean: whether the anchor was found."""
    document = "# Hello There\n\nGeneral.\n"
    async with MarkdownApp(document).run_test() as pilot:
        markdown = pilot.app.query_one(Markdown)
        assert markdown.goto_anchor(anchor) is found


async def test_update_of_document_posts_table_of_content_update_message() -> None:
    """Updating the document should post a TableOfContentsUpdated message."""

    messages: list[str] = []

    class TableOfContentApp(App[None]):
        def compose(self) -> ComposeResult:
            yield Markdown("# One\n\n#Two\n")

        @on(Markdown.TableOfContentsUpdated)
        def log_table_of_content_update(
            self, event: Markdown.TableOfContentsUpdated
        ) -> None:
            nonlocal messages
            messages.append(event.__class__.__name__)

    async with TableOfContentApp().run_test() as pilot:
        assert messages == ["TableOfContentsUpdated"]
        await pilot.app.query_one(Markdown).update("")
        await pilot.pause()
        assert messages == ["TableOfContentsUpdated", "TableOfContentsUpdated"]


async def test_link_in_markdown_table_posts_message_when_clicked():
    """A link inside a markdown table should post a `Markdown.LinkClicked`
    message when clicked.

    Regression test for https://github.com/Textualize/textual/issues/4683
    """

    markdown_table = """\
| Textual Links                                    |
| ------------------------------------------------ |
| [GitHub](https://github.com/textualize/textual/) |
| [Documentation](https://textual.textualize.io/)  |\
"""

    class MarkdownTableApp(App):
        messages = []

        def compose(self) -> ComposeResult:
            yield Markdown(markdown_table)

        @on(Markdown.LinkClicked)
        def log_markdown_link_clicked(
            self,
            event: Markdown.LinkClicked,
        ) -> None:
            self.messages.append(event.__class__.__name__)

    app = MarkdownTableApp()
    async with app.run_test() as pilot:
        await pilot.click(Markdown, offset=(3, 3))
        assert app.messages == ["LinkClicked"]


async def test_markdown_quoting():
    # https://github.com/Textualize/textual/issues/3350
    links = []

    class MyApp(App):
        def compose(self) -> ComposeResult:
            self.md = Markdown(markdown="[tété](tété)")
            yield self.md

        def on_markdown_link_clicked(self, message: Markdown.LinkClicked):
            links.append(message.href)

    app = MyApp()
    async with app.run_test() as pilot:
        await pilot.click(Markdown, offset=(0, 0))
    assert links == ["tété"]



from __future__ import annotations

from textual.widgets import Static


class Tooltip(Static, inherit_css=False):
    DEFAULT_CSS = """
    Tooltip {
        layer: _tooltips;
        margin: 1 2;
        padding: 1 2;
        background: $background;
        width: auto;
        height: auto;
        constrain: inflect;
        max-width: 40;
        display: none;
    }
    """
    DEFAULT_CLASSES = "-textual-system"



from textual.app import App, ComposeResult
from textual.widgets import Button, Static


async def test_text_style_inheritance():
    """Check that changes to text style are inherited in children."""

    class FocusableThing(Static, can_focus=True):
        DEFAULT_CSS = """
        FocusableThing {
            text-style: bold;
        }

        FocusableThing:focus {
            text-style: bold reverse;
        }
        """

        def compose(self) -> ComposeResult:
            yield Static("test", id="child-of-focusable-thing")

    class InheritanceApp(App):
        def compose(self) -> ComposeResult:
            yield Button("button1")
            yield FocusableThing()
            yield Button("button2")

    app = InheritanceApp()
    async with app.run_test() as pilot:
        await pilot.pause()
        child = app.query_one("#child-of-focusable-thing")
        assert child.rich_style.bold
        assert not child.rich_style.reverse
        await pilot.press("tab")
        await pilot.pause()
        assert child.rich_style.bold
        assert child.rich_style.reverse



from __future__ import annotations

from typing import Iterable

import rich.repr
from rich.console import group
from rich.padding import Padding
from rich.table import Table
from rich.text import Text

from textual.color import WHITE, Color

NUMBER_OF_SHADES = 3

# Where no content exists
DEFAULT_DARK_BACKGROUND = "#121212"
# What text usually goes on top off
DEFAULT_DARK_SURFACE = "#1e1e1e"

DEFAULT_LIGHT_SURFACE = "#f5f5f5"
DEFAULT_LIGHT_BACKGROUND = "#efefef"


@rich.repr.auto
class ColorSystem:
    """Defines a standard set of colors and variations for building a UI.

    Primary is the main theme color
    Secondary is a second theme color
    """

    COLOR_NAMES = [
        "primary",
        "secondary",
        "background",
        "primary-background",
        "secondary-background",
        "surface",
        "panel",
        "boost",
        "warning",
        "error",
        "success",
        "accent",
    ]

    def __init__(
        self,
        primary: str,
        secondary: str | None = None,
        warning: str | None = None,
        error: str | None = None,
        success: str | None = None,
        accent: str | None = None,
        foreground: str | None = None,
        background: str | None = None,
        surface: str | None = None,
        panel: str | None = None,
        boost: str | None = None,
        dark: bool = False,
        luminosity_spread: float = 0.15,
        text_alpha: float = 0.95,
    ):
        def parse(color: str | None) -> Color | None:
            if color is None:
                return None
            return Color.parse(color)

        self.primary = Color.parse(primary)
        self.secondary = parse(secondary)
        self.warning = parse(warning)
        self.error = parse(error)
        self.success = parse(success)
        self.accent = parse(accent)
        self.foreground = parse(foreground)
        self.background = parse(background)
        self.surface = parse(surface)
        self.panel = parse(panel)
        self.boost = parse(boost)
        self.dark = dark
        self.luminosity_spread = luminosity_spread
        self.text_alpha = text_alpha

    @property
    def shades(self) -> Iterable[str]:
        """The names of the colors and derived shades."""
        for color in self.COLOR_NAMES:
            for shade_number in range(-NUMBER_OF_SHADES, NUMBER_OF_SHADES + 1):
                if shade_number < 0:
                    yield f"{color}-darken-{abs(shade_number)}"
                elif shade_number > 0:
                    yield f"{color}-lighten-{shade_number}"
                else:
                    yield color

    def generate(self) -> dict[str, str]:
        """Generate a mapping of color name on to a CSS color.

        Returns:
            A mapping of color name on to a CSS-style encoded color
        """

        primary = self.primary
        secondary = self.secondary or primary
        warning = self.warning or primary
        error = self.error or secondary
        success = self.success or secondary
        accent = self.accent or primary

        dark = self.dark
        luminosity_spread = self.luminosity_spread

        if dark:
            background = self.background or Color.parse(DEFAULT_DARK_BACKGROUND)
            surface = self.surface or Color.parse(DEFAULT_DARK_SURFACE)
        else:
            background = self.background or Color.parse(DEFAULT_LIGHT_BACKGROUND)
            surface = self.surface or Color.parse(DEFAULT_LIGHT_SURFACE)

        foreground = self.foreground or (background.inverse)
        boost = self.boost or background.get_contrast_text(1.0).with_alpha(0.04)

        if self.panel is None:
            panel = surface.blend(primary, 0.1, alpha=1)
            if dark:
                panel += boost
        else:
            panel = self.panel

        colors: dict[str, str] = {}

        def luminosity_range(spread) -> Iterable[tuple[str, float]]:
            """Get the range of shades from darken2 to lighten2.

            Returns:
                Iterable of tuples (<SHADE SUFFIX, LUMINOSITY DELTA>)
            """
            luminosity_step = spread / 2
            for n in range(-NUMBER_OF_SHADES, +NUMBER_OF_SHADES + 1):
                if n < 0:
                    label = "-darken"
                elif n > 0:
                    label = "-lighten"
                else:
                    label = ""
                yield (f"{label}{'-' + str(abs(n)) if n else ''}"), n * luminosity_step

        # Color names and color
        COLORS: list[tuple[str, Color]] = [
            ("primary", primary),
            ("secondary", secondary),
            ("primary-background", primary),
            ("secondary-background", secondary),
            ("background", background),
            ("foreground", foreground),
            ("panel", panel),
            ("boost", boost),
            ("surface", surface),
            ("warning", warning),
            ("error", error),
            ("success", success),
            ("accent", accent),
        ]

        # Colors names that have a dark variant
        DARK_SHADES = {"primary-background", "secondary-background"}

        for name, color in COLORS:
            is_dark_shade = dark and name in DARK_SHADES
            spread = luminosity_spread
            for shade_name, luminosity_delta in luminosity_range(spread):
                if color.ansi is not None:
                    colors[f"{name}{shade_name}"] = color.hex
                elif is_dark_shade:
                    dark_background = background.blend(color, 0.15, alpha=1.0)
                    shade_color = dark_background.blend(
                        WHITE, spread + luminosity_delta, alpha=1.0
                    ).clamped
                    colors[f"{name}{shade_name}"] = shade_color.hex
                else:
                    shade_color = color.lighten(luminosity_delta)
                    colors[f"{name}{shade_name}"] = shade_color.hex

        if foreground.ansi is None:
            colors["text"] = "auto 87%"
            colors["text-muted"] = "auto 60%"
            colors["text-disabled"] = "auto 38%"
        else:
            colors["text"] = "ansi_default"
            colors["text-muted"] = "ansi_default"
            colors["text-disabled"] = "ansi_default"

        return colors


def show_design(light: ColorSystem, dark: ColorSystem) -> Table:
    """Generate a renderable to show color systems.

    Args:
        light: Light ColorSystem.
        dark: Dark ColorSystem

    Returns:
        Table showing all colors.
    """

    @group()
    def make_shades(system: ColorSystem):
        colors = system.generate()
        for name in system.shades:
            background = Color.parse(colors[name]).with_alpha(1.0)
            foreground = background + background.get_contrast_text(0.9)

            text = Text(f"${name}")

            yield Padding(text, 1, style=f"{foreground.hex6} on {background.hex6}")

    table = Table(box=None, expand=True)
    table.add_column("Light", justify="center")
    table.add_column("Dark", justify="center")
    table.add_row(make_shades(light), make_shades(dark))
    return table



"""Tests some edits using the keyboard.

All tests in this module should press keys on the keyboard which edit the document,
and check that the document content is updated as expected, as well as the cursor
location.

Note that more extensive testing for editing is done at the Document level.
"""

import pytest

from textual.app import App, ComposeResult
from textual.events import Paste
from textual.widgets import TextArea
from textual.widgets.text_area import Selection

TEXT = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
"""

SIMPLE_TEXT = """\
ABCDE
FGHIJ
KLMNO
PQRST
UVWXY
Z"""


class TextAreaApp(App):
    def compose(self) -> ComposeResult:
        text_area = TextArea.code_editor()
        text_area.load_text(TEXT)
        yield text_area


async def test_single_keypress_printable_character():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        await pilot.press("x")
        assert text_area.text == "x" + TEXT


async def test_single_keypress_enter():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        await pilot.press("enter")
        assert text_area.text == "\n" + TEXT


@pytest.mark.parametrize(
    "content,cursor_column,cursor_destination",
    [
        ("", 0, 4),
        ("x", 0, 4),
        ("x", 1, 4),
        ("xxx", 3, 4),
        ("xxxx", 4, 8),
        ("xxxxx", 5, 8),
        ("xxxxxx", 6, 8),
        ("💩", 1, 3),
        ("💩💩", 2, 6),
    ],
)
async def test_tab_with_spaces_goes_to_tab_stop(
    content, cursor_column, cursor_destination
):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.indent_width = 4
        text_area.load_text(content)
        text_area.cursor_location = (0, cursor_column)

        await pilot.press("tab")

        assert text_area.cursor_location[1] == cursor_destination


async def test_delete_left():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("Hello, world!")
        text_area.move_cursor((0, 6))
        await pilot.press("backspace")
        assert text_area.text == "Hello world!"
        assert text_area.selection == Selection.cursor((0, 5))


async def test_delete_left_start():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("Hello, world!")
        await pilot.press("backspace")
        assert text_area.text == "Hello, world!"
        assert text_area.selection == Selection.cursor((0, 0))


async def test_delete_left_end():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("Hello, world!")
        text_area.move_cursor((0, 13))
        await pilot.press("backspace")
        assert text_area.text == "Hello, world"
        assert text_area.selection == Selection.cursor((0, 12))


@pytest.mark.parametrize(
    "key,selection",
    [
        ("delete", Selection((1, 2), (3, 4))),
        ("delete", Selection((3, 4), (1, 2))),
        ("backspace", Selection((1, 2), (3, 4))),
        ("backspace", Selection((3, 4), (1, 2))),
    ],
)
async def test_deletion_with_non_empty_selection(key, selection):
    """When there's a selection, pressing backspace or delete should delete everything
    that is selected and reset the selection to a cursor at the appropriate location."""
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text(SIMPLE_TEXT)
        text_area.selection = selection
        await pilot.press(key)
        assert text_area.selection == Selection.cursor((1, 2))
        assert (
            text_area.text
            == """\
ABCDE
FGT
UVWXY
Z"""
        )


async def test_delete_right():
    """Pressing 'delete' deletes the character to the right of the cursor."""
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("Hello, world!")
        text_area.move_cursor((0, 13))
        await pilot.press("delete")
        assert text_area.text == "Hello, world!"
        assert text_area.selection == Selection.cursor((0, 13))


async def test_delete_right_end_of_line():
    """Pressing 'delete' at the end of the line merges this line with the line below."""
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("hello\nworld!")
        end_of_line = text_area.get_cursor_line_end_location()
        text_area.move_cursor(end_of_line)
        await pilot.press("delete")
        assert text_area.selection == Selection.cursor((0, 5))
        assert text_area.text == "helloworld!"


@pytest.mark.parametrize(
    "selection,expected_result",
    [
        (Selection.cursor((0, 0)), ""),
        (Selection.cursor((0, 4)), ""),
        (Selection.cursor((0, 10)), ""),
        (Selection((0, 2), (0, 4)), ""),
        (Selection((0, 4), (0, 2)), ""),
    ],
)
async def test_delete_line(selection, expected_result):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("0123456789")
        text_area.selection = selection

        await pilot.press("ctrl+x")

        assert text_area.selection == Selection.cursor((0, 0))
        assert text_area.text == expected_result


@pytest.mark.parametrize(
    "selection,expected_result",
    [
        # Cursors
        (Selection.cursor((0, 0)), "345\n678\n9\n"),
        (Selection.cursor((0, 2)), "345\n678\n9\n"),
        (Selection.cursor((3, 1)), "012\n345\n678\n"),
        (Selection.cursor((4, 0)), "012\n345\n678\n9\n"),
        # Selections
        (Selection((1, 1), (1, 2)), "012\n678\n9\n"),  # non-empty single line selection
        (Selection((1, 2), (2, 1)), "012\n9\n"),  # delete lines selection touches
        (
            Selection((1, 2), (3, 0)),
            "012\n9\n",
        ),  # cursor at column 0 of line 3, should not be deleted!
        (
            Selection((3, 0), (1, 2)),
            "012\n9\n",
        ),  # opposite direction
        (Selection((0, 0), (4, 0)), ""),  # delete all lines
    ],
)
async def test_delete_line_multiline_document(selection, expected_result):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("012\n345\n678\n9\n")
        text_area.selection = selection

        await pilot.press("ctrl+x")

        cursor_row, cursor_column = text_area.cursor_location
        assert text_area.selection == Selection.cursor((cursor_row, cursor_column))
        assert text_area.text == expected_result


@pytest.mark.parametrize(
    "selection,expected_result",
    [
        # Cursors
        (Selection.cursor((0, 0)), ""),
        (Selection.cursor((0, 5)), "01234"),
        (Selection.cursor((0, 9)), "012345678"),
        (Selection.cursor((0, 10)), "0123456789"),
        # Selections
        (Selection((0, 0), (0, 9)), "012345678"),
        (Selection((0, 0), (0, 10)), "0123456789"),
        (Selection((0, 2), (0, 5)), "01234"),
        (Selection((0, 5), (0, 2)), "01"),
    ],
)
async def test_delete_to_end_of_line(selection, expected_result):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("0123456789")
        text_area.selection = selection

        await pilot.press("ctrl+k")

        assert text_area.selection == Selection.cursor(selection.end)
        assert text_area.text == expected_result


@pytest.mark.parametrize(
    "selection,expected_result",
    [
        # Cursors
        (Selection.cursor((0, 0)), "0123456789"),
        (Selection.cursor((0, 5)), "56789"),
        (Selection.cursor((0, 9)), "9"),
        (Selection.cursor((0, 10)), ""),
        # Selections
        (Selection((0, 0), (0, 9)), "9"),
        (Selection((0, 0), (0, 10)), ""),
        (Selection((0, 2), (0, 5)), "56789"),
        (Selection((0, 5), (0, 2)), "23456789"),
    ],
)
async def test_delete_to_start_of_line(selection, expected_result):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("0123456789")
        text_area.selection = selection

        await pilot.press("ctrl+u")

        assert text_area.selection == Selection.cursor((0, 0))
        assert text_area.text == expected_result


@pytest.mark.parametrize(
    "selection,expected_result,final_selection",
    [
        (Selection.cursor((0, 0)), "  012 345 6789", Selection.cursor((0, 0))),
        (Selection.cursor((0, 4)), "  2 345 6789", Selection.cursor((0, 2))),
        (Selection.cursor((0, 5)), "   345 6789", Selection.cursor((0, 2))),
        (
            Selection.cursor((0, 6)),
            "  345 6789",
            Selection.cursor((0, 2)),
        ),
        (Selection.cursor((0, 14)), "  012 345 ", Selection.cursor((0, 10))),
        # When there's a selection and you "delete word left", it just deletes the selection
        (Selection((0, 4), (0, 11)), "  01789", Selection.cursor((0, 4))),
    ],
)
async def test_delete_word_left(selection, expected_result, final_selection):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("  012 345 6789")
        text_area.selection = selection

        await pilot.press("ctrl+w")

        assert text_area.text == expected_result
        assert text_area.selection == final_selection


@pytest.mark.parametrize(
    "selection,expected_result,final_selection",
    [
        (Selection.cursor((0, 0)), "\t012 \t 345\t6789", Selection.cursor((0, 0))),
        (Selection.cursor((0, 4)), "\t \t 345\t6789", Selection.cursor((0, 1))),
        (Selection.cursor((0, 5)), "\t\t 345\t6789", Selection.cursor((0, 1))),
        (
            Selection.cursor((0, 6)),
            "\t 345\t6789",
            Selection.cursor((0, 1)),
        ),
        (Selection.cursor((0, 15)), "\t012 \t 345\t", Selection.cursor((0, 11))),
        # When there's a selection and you "delete word left", it just deletes the selection
        (Selection((0, 4), (0, 11)), "\t0126789", Selection.cursor((0, 4))),
    ],
)
async def test_delete_word_left_with_tabs(selection, expected_result, final_selection):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("\t012 \t 345\t6789")
        text_area.selection = selection

        await pilot.press("ctrl+w")

        assert text_area.text == expected_result
        assert text_area.selection == final_selection


async def test_delete_word_left_to_start_of_line():
    """If no word boundary found when we 'delete word left', then
    the deletion happens to the start of the line."""
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("0123\n   456789")
        text_area.selection = Selection.cursor((1, 3))

        await pilot.press("ctrl+w")

        assert text_area.text == "0123\n456789"
        assert text_area.selection == Selection.cursor((1, 0))


async def test_delete_word_left_at_line_start():
    """If we're at the start of a line and we 'delete word left', the
    line merges with the line above (if possible)."""
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("0123\n   456789")
        text_area.selection = Selection.cursor((1, 0))

        await pilot.press("ctrl+w")

        assert text_area.text == "0123   456789"
        assert text_area.selection == Selection.cursor((0, 4))


@pytest.mark.parametrize(
    "selection,expected_result,final_selection",
    [
        (Selection.cursor((0, 0)), "012 345 6789", Selection.cursor((0, 0))),
        (Selection.cursor((0, 4)), "  01 345 6789", Selection.cursor((0, 4))),
        (Selection.cursor((0, 5)), "  012345 6789", Selection.cursor((0, 5))),
        (Selection.cursor((0, 14)), "  012 345 6789", Selection.cursor((0, 14))),
        # When non-empty selection, "delete word right" just deletes the selection
        (Selection((0, 4), (0, 11)), "  01789", Selection.cursor((0, 4))),
    ],
)
async def test_delete_word_right(selection, expected_result, final_selection):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("  012 345 6789")
        text_area.selection = selection

        await pilot.press("ctrl+f")

        assert text_area.text == expected_result
        assert text_area.selection == final_selection


async def test_delete_word_right_delete_to_end_of_line():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("01234\n56789")
        text_area.selection = Selection.cursor((0, 3))

        await pilot.press("ctrl+f")

        assert text_area.text == "012\n56789"
        assert text_area.selection == Selection.cursor((0, 3))


async def test_delete_word_right_at_end_of_line():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.load_text("01234\n56789")
        text_area.selection = Selection.cursor((0, 5))

        await pilot.press("ctrl+f")

        assert text_area.text == "0123456789"
        assert text_area.selection == Selection.cursor((0, 5))


@pytest.mark.parametrize(
    "binding",
    [
        "enter",
        "backspace",
        "ctrl+u",
        "ctrl+f",
        "ctrl+w",
        "ctrl+k",
        "ctrl+x",
        "space",
        "1",
        "tab",
    ],
)
async def test_edit_read_only_mode_does_nothing(binding):
    """Try out various key-presses and bindings and ensure they don't alter
    the document when read_only=True."""
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.read_only = True
        selection = Selection.cursor((0, 2))
        text_area.selection = selection

        await pilot.press(binding)

        assert text_area.text == TEXT
        assert text_area.selection == selection


@pytest.mark.parametrize(
    "selection",
    [
        Selection(start=(1, 0), end=(3, 0)),
        Selection(start=(3, 0), end=(1, 0)),
    ],
)
async def test_replace_lines_with_fewer_lines(selection):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.text = SIMPLE_TEXT
        text_area.selection = selection

        await pilot.press("a")

        expected_text = """\
ABCDE
aPQRST
UVWXY
Z"""
        assert text_area.text == expected_text
        assert text_area.selection == Selection.cursor((1, 1))


@pytest.mark.parametrize(
    "selection",
    [
        Selection(start=(1, 0), end=(3, 0)),
        Selection(start=(3, 0), end=(1, 0)),
    ],
)
async def test_paste(selection):
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.text = SIMPLE_TEXT
        text_area.selection = selection

        app.post_message(Paste("a"))
        await pilot.pause()

        expected_text = """\
ABCDE
aPQRST
UVWXY
Z"""
        assert text_area.text == expected_text
        assert text_area.selection == Selection.cursor((1, 1))


async def test_paste_read_only_does_nothing():
    app = TextAreaApp()
    async with app.run_test() as pilot:
        text_area = app.query_one(TextArea)
        text_area.read_only = True

        app.post_message(Paste("hello"))
        await pilot.pause()

        assert text_area.text == TEXT  # No change



from textual.app import App, ComposeResult
from textual.widgets import Log

TEXT = """I must not fear.
Fear is the mind-killer.
Fear is the little-death that brings total obliteration.
I will face my fear.
I will permit it to pass over me and through me.
And when it has gone past, I will turn the inner eye to see its path.
Where the fear has gone there will be nothing. Only I will remain."""


class LogApp(App):
    """An app with a simple log."""

    def compose(self) -> ComposeResult:
        yield Log()

    def on_ready(self) -> None:
        log = self.query_one(Log)
        log.write_line("Hello, World!")
        for _ in range(10):
            log.write_line(TEXT)


if __name__ == "__main__":
    app = LogApp()
    app.run()



#!/usr/bin/env python3

from __future__ import annotations

from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.widgets import Static


class NestedAutoApp(App[None]):
    CSS = """
    Screen {
        background: red;
    }

    #my-static-container {
        border: heavy lightgreen;
        background: green;
        height: auto;
        max-height: 10;
    }

    #my-static-wrapper {
        border: heavy lightblue;
        background: blue;
        width: auto;
        height: auto;
    }

    #my-static {
        border: heavy gray;
        background: black;
        width: auto;
        height: auto;
    }
    """
    BINDINGS = [
        ("1", "1", "1"),
        ("2", "2", "2"),
        ("q", "quit", "Quit"),
    ]

    def compose(self) -> ComposeResult:
        self._static = Static("", id="my-static")
        yield VerticalScroll(
            VerticalScroll(
                self._static,
                id="my-static-wrapper",
            ),
            id="my-static-container",
        )

    def action_1(self) -> None:
        self._static.update(
            "\n".join(f"Lorem {i} Ipsum {i} Sit {i}" for i in range(1, 21))
        )

    def action_2(self) -> None:
        self._static.update("JUST ONE LINE")


if __name__ == "__main__":
    app = NestedAutoApp()
    app.run()



from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from dataclasses import dataclass
from functools import partial
from typing import TYPE_CHECKING, Any, Callable, TypeVar

from typing_extensions import Protocol, runtime_checkable

from textual import _time
from textual._callback import invoke
from textual._easing import DEFAULT_EASING, EASING
from textual._types import AnimationLevel, CallbackType
from textual.timer import Timer

if TYPE_CHECKING:
    from textual.app import App

    AnimationKey = tuple[int, str]
    """Animation keys are the id of the object and the attribute being animated."""

EasingFunction = Callable[[float], float]
"""Signature for a function that parametrizes animation speed.

An easing function must map the interval [0, 1] into the interval [0, 1].
"""


class AnimationError(Exception):
    """An issue prevented animation from starting."""


ReturnType = TypeVar("ReturnType")


@runtime_checkable
class Animatable(Protocol):
    """Protocol for objects that can have their intrinsic values animated.

    For example, the transition between two colors can be animated
    because the class [`Color`][textual.color.Color.blend] satisfies this protocol.
    """

    def blend(
        self: ReturnType, destination: ReturnType, factor: float
    ) -> ReturnType:  # pragma: no cover
        ...


class Animation(ABC):
    on_complete: CallbackType | None = None
    """Callback to run after animation completes"""

    @abstractmethod
    def __call__(
        self,
        time: float,
        app_animation_level: AnimationLevel = "full",
    ) -> bool:  # pragma: no cover
        """Call the animation, return a boolean indicating whether animation is in-progress or complete.

        Args:
            time: The current timestamp

        Returns:
            True if the animation has finished, otherwise False.
        """
        raise NotImplementedError("")

    async def invoke_callback(self) -> None:
        """Calls the [`on_complete`][Animation.on_complete] callback if one is provided."""
        if self.on_complete is not None:
            await invoke(self.on_complete)

    @abstractmethod
    async def stop(self, complete: bool = True) -> None:
        """Stop the animation.

        Args:
            complete: Flag to say if the animation should be taken to completion.
        """
        raise NotImplementedError

    def __eq__(self, other: object) -> bool:
        return False


@dataclass
class SimpleAnimation(Animation):
    obj: object
    attribute: str
    start_time: float
    duration: float
    start_value: float | Animatable
    end_value: float | Animatable
    final_value: object
    easing: EasingFunction
    on_complete: CallbackType | None = None
    level: AnimationLevel = "full"
    """Minimum level required for the animation to take place (inclusive)."""

    def __call__(
        self, time: float, app_animation_level: AnimationLevel = "full"
    ) -> bool:
        if (
            self.duration == 0
            or app_animation_level == "none"
            or app_animation_level == "basic"
            and self.level == "full"
        ):
            setattr(self.obj, self.attribute, self.final_value)
            return True

        factor = min(1.0, (time - self.start_time) / self.duration)
        eased_factor = self.easing(factor)

        if factor == 1.0:
            value = self.final_value
        elif isinstance(self.start_value, Animatable):
            assert isinstance(
                self.end_value, Animatable
            ), "end_value must be animatable"
            value = self.start_value.blend(self.end_value, eased_factor)
        else:
            assert isinstance(
                self.start_value, (int, float)
            ), f"`start_value` must be float, not {self.start_value!r}"
            assert isinstance(
                self.end_value, (int, float)
            ), f"`end_value` must be float, not {self.end_value!r}"

            if self.end_value > self.start_value:
                eased_factor = self.easing(factor)
                value = (
                    self.start_value
                    + (self.end_value - self.start_value) * eased_factor
                )
            else:
                eased_factor = 1 - self.easing(factor)
                value = (
                    self.end_value + (self.start_value - self.end_value) * eased_factor
                )
        setattr(self.obj, self.attribute, value)
        return factor >= 1

    async def stop(self, complete: bool = True) -> None:
        """Stop the animation.

        Args:
            complete: Flag to say if the animation should be taken to completion.

        Note:
            [`on_complete`][Animation.on_complete] will be called regardless
            of the value provided for `complete`.
        """
        if complete:
            setattr(self.obj, self.attribute, self.end_value)
        await self.invoke_callback()

    def __eq__(self, other: object) -> bool:
        if isinstance(other, SimpleAnimation):
            return (
                self.final_value == other.final_value
                and self.duration == other.duration
            )
        return False


class BoundAnimator:
    def __init__(self, animator: Animator, obj: object) -> None:
        self._animator = animator
        self._obj = obj

    def __call__(
        self,
        attribute: str,
        value: str | float | Animatable,
        *,
        final_value: object = ...,
        duration: float | None = None,
        speed: float | None = None,
        delay: float = 0.0,
        easing: EasingFunction | str = DEFAULT_EASING,
        on_complete: CallbackType | None = None,
        level: AnimationLevel = "full",
    ) -> None:
        """Animate an attribute.

        Args:
            attribute: Name of the attribute to animate.
            value: The value to animate to.
            final_value: The final value of the animation. Defaults to `value` if not set.
            duration: The duration (in seconds) of the animation.
            speed: The speed of the animation.
            delay: A delay (in seconds) before the animation starts.
            easing: An easing method.
            on_complete: A callable to invoke when the animation is finished.
            level: Minimum level required for the animation to take place (inclusive).
        """
        start_value = getattr(self._obj, attribute)
        if isinstance(value, str) and hasattr(start_value, "parse"):
            # Color and Scalar have a parse method
            # I'm exploiting a coincidence here, but I think this should be a first-class concept
            # TODO: add a `Parsable` protocol
            value = start_value.parse(value)
        easing_function = EASING[easing] if isinstance(easing, str) else easing
        return self._animator.animate(
            self._obj,
            attribute=attribute,
            value=value,
            final_value=final_value,
            duration=duration,
            speed=speed,
            delay=delay,
            easing=easing_function,
            on_complete=on_complete,
            level=level,
        )


class Animator:
    """An object to manage updates to a given attribute over a period of time."""

    def __init__(self, app: App, frames_per_second: int = 60) -> None:
        """Initialise the animator object.

        Args:
            app: The application that owns the animator.
            frames_per_second: The number of frames/second to run the animation at.
        """
        self._animations: dict[AnimationKey, Animation] = {}
        """Dictionary that maps animation keys to the corresponding animation instances."""
        self._scheduled: dict[AnimationKey, Timer] = {}
        """Dictionary of scheduled animations, comprising of their keys and the timer objects."""
        self.app = app
        """The app that owns the animator object."""
        self._timer = Timer(
            app,
            1 / frames_per_second,
            name="Animator",
            callback=self,
            pause=True,
        )
        """The timer that runs the animator."""
        self._idle_event = asyncio.Event()
        """Flag if no animations are currently taking place."""
        self._complete_event = asyncio.Event()
        """Flag if no animations are currently taking place and none are scheduled."""

    async def start(self) -> None:
        """Start the animator task."""
        self._idle_event.set()
        self._complete_event.set()
        self._timer._start()

    async def stop(self) -> None:
        """Stop the animator task."""
        try:
            self._timer.stop()
        except asyncio.CancelledError:
            pass
        finally:
            self._idle_event.set()
            self._complete_event.set()

    def bind(self, obj: object) -> BoundAnimator:
        """Bind the animator to a given object.

        Args:
            obj: The object to bind to.

        Returns:
            The bound animator.
        """
        return BoundAnimator(self, obj)

    def is_being_animated(self, obj: object, attribute: str) -> bool:
        """Does the object/attribute pair have an ongoing or scheduled animation?

        Args:
            obj: An object to check for.
            attribute: The attribute on the object to test for.

        Returns:
            `True` if that attribute is being animated for that object, `False` if not.
        """
        key = (id(obj), attribute)
        return key in self._animations or key in self._scheduled

    def animate(
        self,
        obj: object,
        attribute: str,
        value: Any,
        *,
        final_value: object = ...,
        duration: float | None = None,
        speed: float | None = None,
        easing: EasingFunction | str = DEFAULT_EASING,
        delay: float = 0.0,
        on_complete: CallbackType | None = None,
        level: AnimationLevel = "full",
    ) -> None:
        """Animate an attribute to a new value.

        Args:
            obj: The object containing the attribute.
            attribute: The name of the attribute.
            value: The destination value of the attribute.
            final_value: The final value, or ellipsis if it is the same as ``value``.
            duration: The duration of the animation, or ``None`` to use speed.
            speed: The speed of the animation.
            easing: An easing function.
            delay: Number of seconds to delay the start of the animation by.
            on_complete: Callback to run after the animation completes.
            level: Minimum level required for the animation to take place (inclusive).
        """
        self._record_animation(attribute)
        animate_callback = partial(
            self._animate,
            obj,
            attribute,
            value,
            final_value=final_value,
            duration=duration,
            speed=speed,
            easing=easing,
            on_complete=on_complete,
            level=level,
        )
        if delay:
            self._complete_event.clear()
            self._scheduled[(id(obj), attribute)] = self.app.set_timer(
                delay, animate_callback
            )
        else:
            animate_callback()

    def _record_animation(self, attribute: str) -> None:
        """Called when an attribute is to be animated.

        Args:
            attribute: Attribute being animated.
        """

    def _animate(
        self,
        obj: object,
        attribute: str,
        value: Any,
        *,
        final_value: object = ...,
        duration: float | None = None,
        speed: float | None = None,
        easing: EasingFunction | str = DEFAULT_EASING,
        on_complete: CallbackType | None = None,
        level: AnimationLevel = "full",
    ) -> None:
        """Animate an attribute to a new value.

        Args:
            obj: The object containing the attribute.
            attribute: The name of the attribute.
            value: The destination value of the attribute.
            final_value: The final value, or ellipsis if it is the same as ``value``.
            duration: The duration of the animation, or ``None`` to use speed.
            speed: The speed of the animation.
            easing: An easing function.
            on_complete: Callback to run after the animation completes.
            level: Minimum level required for the animation to take place (inclusive).
        """
        if not hasattr(obj, attribute):
            raise AttributeError(
                f"Can't animate attribute {attribute!r} on {obj!r}; attribute does not exist"
            )
        assert (duration is not None and speed is None) or (
            duration is None and speed is not None
        ), "An Animation should have a duration OR a speed"

        # If an animation is already scheduled for this attribute, unschedule it.
        animation_key = (id(obj), attribute)
        try:
            del self._scheduled[animation_key]
        except KeyError:
            pass

        if final_value is ...:
            final_value = value

        start_time = self._get_time()
        easing_function = EASING[easing] if isinstance(easing, str) else easing
        animation: Animation | None = None

        if hasattr(obj, "__textual_animation__"):
            animation = getattr(obj, "__textual_animation__")(
                attribute,
                getattr(obj, attribute),
                value,
                start_time,
                duration=duration,
                speed=speed,
                easing=easing_function,
                on_complete=on_complete,
                level=level,
            )

        if animation is None:
            if not isinstance(value, (int, float)) and not isinstance(
                value, Animatable
            ):
                raise AnimationError(
                    f"Don't know how to animate {value!r}; "
                    "Can only animate <int>, <float>, or objects with a blend method"
                )

            start_value = getattr(obj, attribute)

            if start_value == value:
                self._animations.pop(animation_key, None)
                return

            if duration is not None:
                animation_duration = duration
            else:
                if hasattr(value, "get_distance_to"):
                    animation_duration = value.get_distance_to(start_value) / (
                        speed or 50
                    )
                else:
                    animation_duration = abs(value - start_value) / (speed or 50)

            animation = SimpleAnimation(
                obj,
                attribute=attribute,
                start_time=start_time,
                duration=animation_duration,
                start_value=start_value,
                end_value=value,
                final_value=final_value,
                easing=easing_function,
                on_complete=(
                    partial(self.app.call_later, on_complete)
                    if on_complete is not None
                    else None
                ),
                level=level,
            )

        assert animation is not None, "animation expected to be non-None"

        current_animation = self._animations.get(animation_key)
        if current_animation is not None and current_animation == animation:
            return

        self._animations[animation_key] = animation
        self._timer.resume()
        self._idle_event.clear()
        self._complete_event.clear()

    async def _stop_scheduled_animation(
        self, key: AnimationKey, complete: bool
    ) -> None:
        """Stop a scheduled animation.

        Args:
            key: The key for the animation to stop.
            complete: Should the animation be moved to its completed state?
        """
        # First off, pull the timer out of the schedule and stop it; it
        # won't be needed.
        try:
            schedule = self._scheduled.pop(key)
        except KeyError:
            return
        schedule.stop()
        # If we've been asked to complete (there's no point in making the
        # animation only to then do nothing with it), and if there was a
        # callback (there will be, but this just keeps type checkers happy
        # really)...
        if complete and schedule._callback is not None:
            # ...invoke it to get the animator created and in the running
            # animations. Yes, this does mean that a stopped scheduled
            # animation will start running early...
            await invoke(schedule._callback)
            # ...but only so we can call on it to run right to the very end
            # right away.
            await self._stop_running_animation(key, complete)

    async def _stop_running_animation(self, key: AnimationKey, complete: bool) -> None:
        """Stop a running animation.

        Args:
            key: The key for the animation to stop.
            complete: Should the animation be moved to its completed state?
        """
        try:
            animation = self._animations.pop(key)
        except KeyError:
            return
        await animation.stop(complete)

    async def stop_animation(
        self, obj: object, attribute: str, complete: bool = True
    ) -> None:
        """Stop an animation on an attribute.

        Args:
            obj: The object containing the attribute.
            attribute: The name of the attribute.
            complete: Should the animation be set to its final value?

        Note:
            If there is no animation scheduled or running, this is a no-op.
        """
        key = (id(obj), attribute)
        if key in self._scheduled:
            await self._stop_scheduled_animation(key, complete)
        elif key in self._animations:
            await self._stop_running_animation(key, complete)

    def force_stop_animation(self, obj: object, attribute: str) -> None:
        """Force stop an animation on an attribute. This will immediately stop the animation,
        without running any associated callbacks, setting the attribute to its final value.

        Args:
            obj: The object containing the attribute.
            attribute: The name of the attribute.

        Note:
            If there is no animation scheduled or running, this is a no-op.
        """
        from textual.css.scalar_animation import ScalarAnimation

        animation_key = (id(obj), attribute)
        try:
            animation = self._animations.pop(animation_key)
        except KeyError:
            return

        if isinstance(animation, SimpleAnimation):
            setattr(obj, attribute, animation.end_value)
        elif isinstance(animation, ScalarAnimation):
            setattr(obj, attribute, animation.final_value)

        if animation.on_complete is not None:
            animation.on_complete()

    def __call__(self) -> None:
        if not self._animations:
            self._timer.pause()
            self._idle_event.set()
            if not self._scheduled:
                self._complete_event.set()
        else:
            app_animation_level = self.app.animation_level
            animation_time = self._get_time()
            animation_keys = list(self._animations.keys())
            for animation_key in animation_keys:
                animation = self._animations[animation_key]
                animation_complete = animation(animation_time, app_animation_level)
                if animation_complete:
                    del self._animations[animation_key]
                    if animation.on_complete is not None:
                        animation.on_complete()

    def _get_time(self) -> float:
        """Get the current wall clock time, via the internal Timer.

        Returns:
            The wall clock time.
        """
        # N.B. We could remove this method and always call `self._timer.get_time()` internally,
        # but it's handy to have in mocking situations.
        return _time.get_time()

    async def wait_for_idle(self) -> None:
        """Wait for any animations to complete."""
        await self._idle_event.wait()

    async def wait_until_complete(self) -> None:
        """Wait for any current and scheduled animations to complete."""
        await self._complete_event.wait()



from textual.app import App
from textual.screen import Screen
from textual.widgets import TextArea

TEXT = """\
foo
bar
baz
"""


class AltScreen(Screen[None]):
    def compose(self):
        yield TextArea(TEXT)


class TABug(App[None]):
    def on_mount(self):
        self.push_screen(AltScreen())


if __name__ == "__main__":
    app = TABug()
    app.run()



"""Provides a simple Label widget."""

from textual.widgets._static import Static


class Label(Static):
    """A simple label widget for displaying text-oriented renderables."""

    DEFAULT_CSS = """
    Label {
        width: auto;
        height: auto;
        min-height: 1;
    }
    """



from textual import events
from textual.app import App, ComposeResult
from textual.widgets import RichLog, Static


class Ball(Static):
    pass


class MouseApp(App):
    CSS_PATH = "mouse01.tcss"

    def compose(self) -> ComposeResult:
        yield RichLog()
        yield Ball("Textual")

    def on_mouse_move(self, event: events.MouseMove) -> None:
        self.screen.query_one(RichLog).write(event)
        self.query_one(Ball).offset = event.screen_offset - (8, 2)


if __name__ == "__main__":
    app = MouseApp()
    app.run()



"""
Test that generic animations can be disabled.
"""

from textual.app import App, ComposeResult
from textual.color import Color
from textual.widgets import Label


class SingleLabelApp(App[None]):
    """Single label whose background colour we'll animate."""

    CSS = """
        Label {
            background: red;
        }
    """

    def compose(self) -> ComposeResult:
        yield Label()


async def test_style_animations_via_animate_work_on_full() -> None:
    app = SingleLabelApp()
    app.animation_level = "full"

    async with app.run_test():
        label = app.query_one(Label)
        # Sanity check.
        assert label.styles.background == Color.parse("red")
        animator = app.animator
        # Freeze time at 0 before triggering the animation.
        animator._get_time = lambda *_: 0
        label.styles.animate("background", "blue", duration=1)
        assert len(animator._animations) > 0  # Sanity check.
        # Freeze time around the animation midpoint.
        animator._get_time = lambda *_: 0.5
        # Move to the next frame.
        animator()
        # The animation shouldn't have completed.
        assert label.styles.background != Color.parse("red")
        assert label.styles.background != Color.parse("blue")


async def test_style_animations_via_animate_are_disabled_on_basic() -> None:
    app = SingleLabelApp()
    app.animation_level = "basic"

    async with app.run_test():
        label = app.query_one(Label)
        # Sanity check.
        assert label.styles.background == Color.parse("red")
        animator = app.animator
        # Freeze time at 0 before triggering the animation.
        animator._get_time = lambda *_: 0
        label.styles.animate("background", "blue", duration=1)
        assert len(animator._animations) > 0  # Sanity check.
        # Freeze time after the animation start and before animation end.
        animator._get_time = lambda *_: 0.01
        # Move to the next frame.
        animator()
        # The animation should have completed.
        assert label.styles.background == Color.parse("blue")


async def test_style_animations_via_animate_are_disabled_on_none() -> None:
    app = SingleLabelApp()
    app.animation_level = "none"

    async with app.run_test():
        label = app.query_one(Label)
        # Sanity check.
        assert label.styles.background == Color.parse("red")
        animator = app.animator
        # Freeze time at 0 before triggering the animation.
        animator._get_time = lambda *_: 0
        label.styles.animate("background", "blue", duration=1)
        assert len(animator._animations) > 0  # Sanity check.
        # Freeze time after the animation start and before animation end.
        animator._get_time = lambda *_: 0.01
        # Move to the next frame.
        animator()
        # The animation should have completed.
        assert label.styles.background == Color.parse("blue")


class LabelWithTransitionsApp(App[None]):
    """Single label whose background is set to animate with TCSS."""

    CSS = """
        Label {
            background: red;
            transition: background 1s;
        }

        Label.blue-bg {
            background: blue;
        }
    """

    def compose(self) -> ComposeResult:
        yield Label()


async def test_style_animations_via_transition_work_on_full() -> None:
    app = LabelWithTransitionsApp()
    app.animation_level = "full"

    async with app.run_test():
        label = app.query_one(Label)
        # Sanity check.
        assert label.styles.background == Color.parse("red")
        animator = app.animator
        # Free time at 0 before triggering the animation.
        animator._get_time = lambda *_: 0
        label.add_class("blue-bg")
        assert len(animator._animations) > 0  # Sanity check.
        # Freeze time in the middle of the animation.
        animator._get_time = lambda *_: 0.5
        animator()
        # The animation should be undergoing.
        assert label.styles.background != Color.parse("red")
        assert label.styles.background != Color.parse("blue")


async def test_style_animations_via_transition_are_disabled_on_basic() -> None:
    app = LabelWithTransitionsApp()
    app.animation_level = "basic"

    async with app.run_test():
        label = app.query_one(Label)
        # Sanity check.
        assert label.styles.background == Color.parse("red")
        animator = app.animator
        # Free time at 0 before triggering the animation.
        animator._get_time = lambda *_: 0
        label.add_class("blue-bg")
        assert len(animator._animations) > 0  # Sanity check.
        # Freeze time after the animation start and before animation end.
        animator._get_time = lambda *_: 0.01
        animator()
        # The animation should have completed.
        assert label.styles.background == Color.parse("blue")


async def test_style_animations_via_transition_are_disabled_on_none() -> None:
    app = LabelWithTransitionsApp()
    app.animation_level = "none"

    async with app.run_test():
        label = app.query_one(Label)
        # Sanity check.
        assert label.styles.background == Color.parse("red")
        animator = app.animator
        # Free time at 0 before triggering the animation.
        animator._get_time = lambda *_: 0
        label.add_class("blue-bg")
        assert len(animator._animations) > 0  # Sanity check.
        # Freeze time after the animation start and before animation end.
        animator._get_time = lambda *_: 0.01
        animator()
        # The animation should have completed.
        assert label.styles.background == Color.parse("blue")



from __future__ import annotations

from textual.app import App, ComposeResult
from textual.containers import Container
from textual.geometry import clamp
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Input, Label, Switch


class BitSwitch(Widget):
    """A Switch with a numeric label above it."""

    DEFAULT_CSS = """
    BitSwitch {
        layout: vertical;
        width: auto;
        height: auto;
    }
    BitSwitch > Label {
        text-align: center;
        width: 100%;
    }
    """

    class BitChanged(Message):
        """Sent when the 'bit' changes."""

        def __init__(self, bit: int, value: bool) -> None:
            super().__init__()
            self.bit = bit
            self.value = value

    value = reactive(0)

    def __init__(self, bit: int) -> None:
        self.bit = bit
        super().__init__()

    def compose(self) -> ComposeResult:
        yield Label(str(self.bit))
        yield Switch()

    def watch_value(self, value: bool) -> None:  # (1)!
        """When the value changes we want to set the switch accordingly."""
        self.query_one(Switch).value = value

    def on_switch_changed(self, event: Switch.Changed) -> None:
        """When the switch changes, notify the parent via a message."""
        event.stop()
        self.value = event.value
        self.post_message(self.BitChanged(self.bit, event.value))


class ByteInput(Widget):
    """A compound widget with 8 switches."""

    DEFAULT_CSS = """
    ByteInput {
        width: auto;
        height: auto;
        border: blank;
        layout: horizontal;
    }
    ByteInput:focus-within {
        border: heavy $secondary;
    }
    """

    def compose(self) -> ComposeResult:
        for bit in reversed(range(8)):
            yield BitSwitch(bit)


class ByteEditor(Widget):
    DEFAULT_CSS = """
    ByteEditor > Container {
        height: 1fr;
        align: center middle;
    }
    ByteEditor > Container.top {
        background: $boost;
    }
    ByteEditor Input {
        width: 16;
    }
    """

    value = reactive(0)

    def validate_value(self, value: int) -> int:  # (2)!
        """Ensure value is between 0 and 255."""
        return clamp(value, 0, 255)

    def compose(self) -> ComposeResult:
        with Container(classes="top"):
            yield Input(placeholder="byte")
        with Container():
            yield ByteInput()

    def on_bit_switch_bit_changed(self, event: BitSwitch.BitChanged) -> None:
        """When a switch changes, update the value."""
        value = 0
        for switch in self.query(BitSwitch):
            value |= switch.value << switch.bit
        self.query_one(Input).value = str(value)

    def on_input_changed(self, event: Input.Changed) -> None:  # (3)!
        """When the text changes, set the value of the byte."""
        try:
            self.value = int(event.value or "0")
        except ValueError:
            pass

    def watch_value(self, value: int) -> None:  # (4)!
        """When self.value changes, update switches."""
        for switch in self.query(BitSwitch):
            with switch.prevent(BitSwitch.BitChanged):  # (5)!
                switch.value = bool(value & (1 << switch.bit))  # (6)!


class ByteInputApp(App):
    def compose(self) -> ComposeResult:
        yield ByteEditor()


if __name__ == "__main__":
    app = ByteInputApp()
    app.run()



from __future__ import annotations

import pytest

from textual.dom import DOMNode
from textual.suggester import Suggester, SuggestionReady


class FillSuggester(Suggester):
    async def get_suggestion(self, value: str):
        if len(value) <= 10:
            return f"{value:x<10}"


class LogListNode(DOMNode):
    def __init__(self, log_list: list[tuple[str, str]]) -> None:
        self.log_list = log_list

    def post_message(self, message: SuggestionReady):
        # We hijack post_message so we can intercept messages without creating a full app.
        self.log_list.append((message.suggestion, message.value))


async def test_cache_on():
    log = []

    class MySuggester(Suggester):
        async def get_suggestion(self, value: str):
            log.append(value)
            return value

    suggester = MySuggester(use_cache=True)
    await suggester._get_suggestion(DOMNode(), "hello")
    assert log == ["hello"]
    await suggester._get_suggestion(DOMNode(), "hello")
    assert log == ["hello"]


async def test_cache_off():
    log = []

    class MySuggester(Suggester):
        async def get_suggestion(self, value: str):
            log.append(value)
            return value

    suggester = MySuggester(use_cache=False)
    await suggester._get_suggestion(DOMNode(), "hello")
    assert log == ["hello"]
    await suggester._get_suggestion(DOMNode(), "hello")
    assert log == ["hello", "hello"]


async def test_suggestion_ready_message():
    log = []
    suggester = FillSuggester()
    await suggester._get_suggestion(LogListNode(log), "hello")
    assert log == [("helloxxxxx", "hello")]
    await suggester._get_suggestion(LogListNode(log), "world")
    assert log == [("helloxxxxx", "hello"), ("worldxxxxx", "world")]


async def test_no_message_if_no_suggestion():
    log = []
    suggester = FillSuggester()
    await suggester._get_suggestion(LogListNode(log), "this is a longer string")
    assert log == []


async def test_suggestion_ready_message_on_cache_hit():
    log = []
    suggester = FillSuggester(use_cache=True)
    await suggester._get_suggestion(LogListNode(log), "hello")
    assert log == [("helloxxxxx", "hello")]
    await suggester._get_suggestion(LogListNode(log), "hello")
    assert log == [("helloxxxxx", "hello"), ("helloxxxxx", "hello")]


@pytest.mark.parametrize(
    "value",
    [
        "hello",
        "HELLO",
        "HeLlO",
        "Hello",
        "hELLO",
    ],
)
async def test_case_insensitive_suggestions(value):
    class MySuggester(Suggester):
        async def get_suggestion(self, value: str):
            assert "hello" == value

    suggester = MySuggester(use_cache=False, case_sensitive=False)
    await suggester._get_suggestion(DOMNode(), value)


async def test_case_insensitive_cache_hits():
    count = 0

    class MySuggester(Suggester):
        async def get_suggestion(self, value: str):
            nonlocal count
            count += 1
            return value + "abc"

    suggester = MySuggester(use_cache=True, case_sensitive=False)
    hellos = ["hello", "HELLO", "HeLlO", "Hello", "hELLO"]
    for hello in hellos:
        await suggester._get_suggestion(DOMNode(), hello)
    assert count == 1



"""Provides a check box widget."""

from __future__ import annotations

from textual.widgets._toggle_button import ToggleButton


class Checkbox(ToggleButton):
    """A check box widget that represents a boolean value."""

    class Changed(ToggleButton.Changed):
        """Posted when the value of the checkbox changes.

        This message can be handled using an `on_checkbox_changed` method.
        """

        @property
        def checkbox(self) -> Checkbox:
            """The checkbox that was changed."""
            assert isinstance(self._toggle_button, Checkbox)
            return self._toggle_button

        @property
        def control(self) -> Checkbox:
            """An alias for [Changed.checkbox][textual.widgets.Checkbox.Changed.checkbox]."""
            return self.checkbox



import pytest

from textual._spatial_map import SpatialMap
from textual.geometry import Region


@pytest.mark.parametrize(
    "region,grid",
    [
        (
            Region(0, 0, 10, 10),
            [
                (0, 0),
            ],
        ),
        (
            Region(10, 10, 10, 10),
            [
                (1, 1),
            ],
        ),
        (
            Region(0, 0, 11, 11),
            [(0, 0), (0, 1), (1, 0), (1, 1)],
        ),
        (
            Region(5, 5, 15, 3),
            [(0, 0), (1, 0)],
        ),
        (
            Region(5, 5, 2, 15),
            [(0, 0), (0, 1)],
        ),
    ],
)
def test_region_to_grid(region, grid):
    spatial_map = SpatialMap(10, 10)

    assert list(spatial_map._region_to_grid_coordinates(region)) == grid


def test_get_values_in_region() -> None:
    spatial_map: SpatialMap[str] = SpatialMap(20, 10)

    spatial_map.insert(
        [
            (Region(10, 5, 5, 5), False, False, "foo"),
            (Region(5, 20, 5, 5), False, False, "bar"),
            (Region(0, 0, 40, 1), True, False, "title"),
        ]
    )

    assert spatial_map.get_values_in_region(Region(0, 0, 10, 5)) == [
        "title",
        "foo",
    ]
    assert spatial_map.get_values_in_region(Region(0, 1, 10, 5)) == ["title", "foo"]
    assert spatial_map.get_values_in_region(Region(0, 10, 10, 5)) == ["title"]
    assert spatial_map.get_values_in_region(Region(0, 20, 10, 5)) == ["title", "bar"]
    assert spatial_map.get_values_in_region(Region(5, 5, 50, 50)) == [
        "title",
        "foo",
        "bar",
    ]



from textual.app import App, ComposeResult
from textual.containers import ScrollableContainer
from textual.widgets import Button, Footer, Header, Static


class TimeDisplay(Static):
    """A widget to display elapsed time."""


class Stopwatch(Static):
    """A stopwatch widget."""

    def compose(self) -> ComposeResult:
        """Create child widgets of a stopwatch."""
        yield Button("Start", id="start", variant="success")
        yield Button("Stop", id="stop", variant="error")
        yield Button("Reset", id="reset")
        yield TimeDisplay("00:00:00.00")


class StopwatchApp(App):
    """A Textual app to manage stopwatches."""

    BINDINGS = [("d", "toggle_dark", "Toggle dark mode")]

    def compose(self) -> ComposeResult:
        """Create child widgets for the app."""
        yield Header()
        yield Footer()
        yield ScrollableContainer(Stopwatch(), Stopwatch(), Stopwatch())

    def action_toggle_dark(self) -> None:
        """An action to toggle dark mode."""
        self.dark = not self.dark


if __name__ == "__main__":
    app = StopwatchApp()
    app.run()



from contextlib import nullcontext as does_not_raise

import pytest

from textual.color import Color
from textual.css.stylesheet import CssSource, Stylesheet, StylesheetParseError
from textual.css.tokenizer import TokenError
from textual.dom import DOMNode
from textual.geometry import Spacing
from textual.widget import Widget


def _make_user_stylesheet(css: str) -> Stylesheet:
    stylesheet = Stylesheet()
    stylesheet.source["test.tcss"] = CssSource(css, is_defaults=False)
    stylesheet.parse()
    return stylesheet


def test_stylesheet_apply_highest_specificity_wins():
    """#ids have higher specificity than .classes"""
    css = "#id {color: red;} .class {color: blue;}"
    stylesheet = _make_user_stylesheet(css)
    node = DOMNode(classes="class", id="id")
    stylesheet.apply(node)

    assert node.styles.color == Color(255, 0, 0)


def test_stylesheet_apply_doesnt_override_defaults():
    css = "#id {color: red;}"
    stylesheet = _make_user_stylesheet(css)
    node = DOMNode(id="id")
    stylesheet.apply(node)

    assert node.styles.margin == Spacing.all(0)
    assert node.styles.box_sizing == "border-box"


def test_stylesheet_apply_highest_specificity_wins_multiple_classes():
    """When we use two selectors containing only classes, then the selector
    `.b.c` has greater specificity than the selector `.a`"""
    css = ".b.c {background: blue;} .a {background: red; color: lime;}"
    stylesheet = _make_user_stylesheet(css)
    node = DOMNode(classes="a b c")
    stylesheet.apply(node)

    assert node.styles.background == Color(0, 0, 255)
    assert node.styles.color == Color(0, 255, 0)


def test_stylesheet_many_classes_dont_overrule_id():
    """#id is further to the left in the specificity tuple than class, and
    a selector containing multiple classes cannot take priority over even a
    single class."""
    css = "#id {color: red;} .a.b.c.d {color: blue;}"
    stylesheet = _make_user_stylesheet(css)
    node = DOMNode(classes="a b c d", id="id")
    stylesheet.apply(node)

    assert node.styles.color == Color(255, 0, 0)


def test_stylesheet_last_rule_wins_when_same_rule_twice_in_one_ruleset():
    css = "#id {color: red; color: blue;}"
    stylesheet = _make_user_stylesheet(css)
    node = DOMNode(id="id")
    stylesheet.apply(node)

    assert node.styles.color == Color(0, 0, 255)


def test_stylesheet_rulesets_merged_for_duplicate_selectors():
    css = "#id {color: red; background: lime;} #id {color:blue;}"
    stylesheet = _make_user_stylesheet(css)
    node = DOMNode(id="id")
    stylesheet.apply(node)

    assert node.styles.color == Color(0, 0, 255)
    assert node.styles.background == Color(0, 255, 0)


def test_stylesheet_apply_takes_final_rule_in_specificity_clash():
    """.a and .b both contain background and have same specificity, so .b wins
    since it was declared last - the background should be blue."""
    css = ".a {background: red; color: lime;} .b {background: blue;}"
    stylesheet = _make_user_stylesheet(css)
    node = DOMNode(classes="a b", id="c")
    stylesheet.apply(node)

    assert node.styles.color == Color(0, 255, 0)  # color: lime
    assert node.styles.background == Color(0, 0, 255)  # background: blue


def test_stylesheet_apply_empty_rulesets():
    """Ensure that we don't crash when working with empty rulesets"""
    css = ".a {} .b {}"
    stylesheet = _make_user_stylesheet(css)
    node = DOMNode(classes="a b")
    stylesheet.apply(node)


def test_stylesheet_apply_user_css_over_widget_css():
    user_css = ".a {color: red; tint: yellow;}"

    class MyWidget(Widget):
        DEFAULT_CSS = ".a {color: blue !important; background: lime;}"

    node = MyWidget()
    node.add_class("a")

    stylesheet = _make_user_stylesheet(user_css)
    stylesheet.add_source(
        MyWidget.DEFAULT_CSS, "widget.py:MyWidget", is_default_css=True
    )
    stylesheet.apply(node)

    # The node is red because user CSS overrides Widget.DEFAULT_CSS
    assert node.styles.color == Color(255, 0, 0)
    # The background colour defined in the Widget still applies, since user CSS doesn't override it
    assert node.styles.background == Color(0, 255, 0)
    # As expected, the tint colour is yellow, since there's no competition between user or widget CSS
    assert node.styles.tint == Color(255, 255, 0)


@pytest.mark.parametrize(
    "css_value,expectation,expected_color",
    [
        # Valid values:
        ["transparent", does_not_raise(), Color(0, 0, 0, 0)],
        ["ansi_red", does_not_raise(), Color(128, 0, 0, ansi=1)],
        ["ansi_bright_magenta", does_not_raise(), Color(255, 0, 255, ansi=13)],
        ["red", does_not_raise(), Color(255, 0, 0)],
        ["lime", does_not_raise(), Color(0, 255, 0)],
        ["coral", does_not_raise(), Color(255, 127, 80)],
        ["aqua", does_not_raise(), Color(0, 255, 255)],
        ["deepskyblue", does_not_raise(), Color(0, 191, 255)],
        ["rebeccapurple", does_not_raise(), Color(102, 51, 153)],
        ["#ffcc00", does_not_raise(), Color(255, 204, 0)],
        ["#ffcc0033", does_not_raise(), Color(255, 204, 0, 0.2)],
        ["rgb(200,90,30)", does_not_raise(), Color(200, 90, 30)],
        ["rgba(200,90,30,0.3)", does_not_raise(), Color(200, 90, 30, 0.3)],
        # Some invalid ones:
        ["coffee", pytest.raises(StylesheetParseError), None],  # invalid color name
        ["ansi_dark_cyan", pytest.raises(StylesheetParseError), None],
        ["red 4", pytest.raises(StylesheetParseError), None],  # space in it
        ["1", pytest.raises(StylesheetParseError), None],  # invalid value
        ["()", pytest.raises(TokenError), None],  # invalid tokens
    ],
)
def test_color_property_parsing(css_value, expectation, expected_color):
    stylesheet = Stylesheet()
    css = """
    * {
      background: ${COLOR};
    }
    """.replace(
        "${COLOR}", css_value
    )

    with expectation:
        stylesheet.add_source(css)
        stylesheet.parse()

    if expected_color:
        css_rule = stylesheet.rules[0]
        assert css_rule.styles.background == expected_color


@pytest.mark.parametrize(
    "css_property_name,expected_property_name_suggestion",
    [
        ["backgroundu", "background"],
        ["bckgroundu", "background"],
        ["ofset-x", "offset-x"],
        ["ofst_y", "offset-y"],
        ["colr", "color"],
        ["colour", "color"],
        ["wdth", "width"],
        ["wth", "width"],
        ["wh", None],
        ["xkcd", None],
    ],
)
def test_did_you_mean_for_css_property_names(
    css_property_name: str, expected_property_name_suggestion
):
    stylesheet = Stylesheet()
    css = """
    * {
      border: blue;
      ${PROPERTY}: red;
    }
    """.replace(
        "${PROPERTY}", css_property_name
    )

    stylesheet.add_source(css)
    with pytest.raises(StylesheetParseError) as err:
        stylesheet.parse()

    _, help_text = err.value.errors.rules[0].errors[0]  # type: Any, HelpText
    displayed_css_property_name = css_property_name.replace("_", "-")
    expected_summary = f"Invalid CSS property {displayed_css_property_name!r}"
    if expected_property_name_suggestion:
        expected_summary += f". Did you mean '{expected_property_name_suggestion}'?"
    assert help_text.summary == expected_summary


@pytest.mark.parametrize(
    "css_property_name,expected_property_name_suggestion",
    [
        ["backgroundu", "background"],
        ["bckgroundu", "background"],
        ["ofset-x", "offset-x"],
        ["ofst_y", "offset-y"],
        ["colr", "color"],
        ["colour", "color"],
        ["wdth", "width"],
        ["wth", "width"],
        ["wh", None],
        ["xkcd", None],
    ],
)
def test_did_you_mean_for_property_names_in_nested_css(
    css_property_name: str, expected_property_name_suggestion: "str | None"
) -> None:
    """Test that we get nice errors with mistyped declaractions in nested CSS.

    When implementing pseudo-class support in nested TCSS
    (https://github.com/Textualize/textual/issues/4039), the first iterations didn't
    preserve this so we add these tests to make sure we don't take this feature away
    unintentionally.
    """
    stylesheet = Stylesheet()
    css = """
    Screen {
        * {
            border: blue;
            ${PROPERTY}: red;
        }
    }
    """.replace(
        "${PROPERTY}", css_property_name
    )

    stylesheet.add_source(css)
    with pytest.raises(StylesheetParseError) as err:
        stylesheet.parse()

    _, help_text = err.value.errors.rules[1].errors[0]
    displayed_css_property_name = css_property_name.replace("_", "-")
    expected_summary = f"Invalid CSS property {displayed_css_property_name!r}"
    if expected_property_name_suggestion:
        expected_summary += f". Did you mean '{expected_property_name_suggestion}'?"
    assert help_text.summary == expected_summary


@pytest.mark.parametrize(
    "css_property_name,css_property_value,expected_color_suggestion",
    [
        ["color", "blu", "blue"],
        ["background", "chartruse", "chartreuse"],
        ["tint", "ansi_whi", "ansi_white"],
        ["scrollbar-color", "transprnt", "transparent"],
        ["color", "xkcd", None],
    ],
)
def test_did_you_mean_for_color_names(
    css_property_name: str, css_property_value: str, expected_color_suggestion
):
    stylesheet = Stylesheet()
    css = """
    * {
      border: blue;
      ${PROPERTY}: ${VALUE};
    }
    """.replace(
        "${PROPERTY}", css_property_name
    ).replace(
        "${VALUE}", css_property_value
    )

    stylesheet.add_source(css)
    with pytest.raises(StylesheetParseError) as err:
        stylesheet.parse()

    _, help_text = err.value.errors.rules[0].errors[0]  # type: Any, HelpText
    displayed_css_property_name = css_property_name.replace("_", "-")
    expected_error_summary = (
        f"Invalid value for the [i]{displayed_css_property_name}[/] property"
    )

    if expected_color_suggestion is not None:
        expected_error_summary += f". Did you mean '{expected_color_suggestion}'?"

    assert help_text.summary == expected_error_summary



from textual.eta import ETA


def test_basics() -> None:
    eta = ETA()
    eta.add_sample(1.0, 1.0)
    assert eta.first_sample == (0, 0)
    assert eta.last_sample == (1.0, 1.0)
    assert len(eta._samples) == 2
    repr(eta)


def test_speed() -> None:
    eta = ETA()
    # One sample is not enough to determine speed
    assert eta.speed is None
    eta.add_sample(1.0, 0.5)
    assert eta.speed == 0.5

    # Check reset
    eta.reset()
    assert eta.speed is None
    eta.add_sample(0.0, 0.0)
    assert eta.speed is None
    eta.add_sample(1.0, 0.5)
    assert eta.speed == 0.5

    # Check backwards
    eta.add_sample(2.0, 0.0)
    assert eta.speed is None
    eta.add_sample(3.0, 1.0)
    assert eta.speed == 1.0


def test_get_progress_at() -> None:
    eta = ETA()
    eta.add_sample(1, 2)
    # Check interpolation works
    assert eta._get_progress_at(0) == (0, 0)
    assert eta._get_progress_at(0.1) == (0.1, 0.2)
    assert eta._get_progress_at(0.5) == (0.5, 1.0)


def test_eta():
    eta = ETA(estimation_period=2, extrapolate_period=5)
    eta.add_sample(1, 0.1)
    assert eta.speed == 0.1
    assert eta.get_eta(1) == 9
    assert eta.get_eta(2) == 8
    assert eta.get_eta(3) == 7
    assert eta.get_eta(4) == 6
    assert eta.get_eta(5) == 5
    # After 5 seconds (extrapolate_period), eta won't update
    assert eta.get_eta(6) == 4
    assert eta.get_eta(7) == 4
    assert eta.get_eta(8) == 4



from __future__ import annotations

from fractions import Fraction

from textual.box_model import BoxModel
from textual.geometry import Size, Spacing
from textual.widget import Widget


def test_content_box():
    one = Fraction(1)

    class TestWidget(Widget):
        def get_content_width(self, container: Size, parent: Size) -> int:
            assert False, "must not be called"

        def get_content_height(self, container: Size, parent: Size) -> int:
            assert False, "must not be called"

    widget = TestWidget()

    # border-box is default
    assert widget.styles.box_sizing == "border-box"

    widget.styles.width = 10
    widget.styles.height = 8
    widget.styles.padding = 1
    widget.styles.border = ("solid", "red")

    box_model = widget._get_box_model(
        Size(60, 20),
        Size(80, 24),
        one,
        one,
    )
    # Size should be inclusive of padding / border
    assert box_model == BoxModel(Fraction(10), Fraction(8), Spacing(0, 0, 0, 0))

    # Switch to content-box
    widget.styles.box_sizing = "content-box"

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    # width and height have added padding / border to accommodate content
    assert box_model == BoxModel(Fraction(14), Fraction(12), Spacing(0, 0, 0, 0))


def test_width():
    """Test width settings."""

    one = Fraction(1)

    class TestWidget(Widget):
        def get_content_width(self, container: Size, parent: Size) -> int:
            return 10

        def get_content_height(self, container: Size, parent: Size, width: int) -> int:
            return 10

    widget = TestWidget()
    styles = widget.styles
    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(60), Fraction(20), Spacing(0, 0, 0, 0))

    # Add a margin and check that it is reported
    styles.margin = (1, 2, 3, 4)

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(54), Fraction(16), Spacing(1, 2, 3, 4))

    # Set width to auto-detect
    styles.width = "auto"

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    # Setting width to auto should call get_auto_width
    assert box_model == BoxModel(Fraction(10), Fraction(16), Spacing(1, 2, 3, 4))

    # Set width to 100 vw which should make it the width of the parent
    styles.width = "100vw"

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(80), Fraction(16), Spacing(1, 2, 3, 4))

    # Set the width to 100% should make it fill the container size
    styles.width = "100%"

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(54), Fraction(16), Spacing(1, 2, 3, 4))

    styles.width = "100vw"
    styles.max_width = "50%"

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(27), Fraction(16), Spacing(1, 2, 3, 4))


def test_height():
    """Test height settings."""

    one = Fraction(1)

    class TestWidget(Widget):
        def get_content_width(self, container: Size, parent: Size) -> int:
            return 10

        def get_content_height(self, container: Size, parent: Size, width: int) -> int:
            return 10

    widget = TestWidget()
    styles = widget.styles

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(60), Fraction(20), Spacing(0, 0, 0, 0))

    # Add a margin and check that it is reported
    styles.margin = (1, 2, 3, 4)

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(54), Fraction(16), Spacing(1, 2, 3, 4))

    # Set height to 100 vw which should make it the height of the parent
    styles.height = "100vh"

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(54), Fraction(24), Spacing(1, 2, 3, 4))

    # Set the height to 100% should make it fill the container size
    styles.height = "100%"

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(54), Fraction(16), Spacing(1, 2, 3, 4))

    styles.height = "auto"
    styles.margin = 2

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    print(box_model)
    assert box_model == BoxModel(Fraction(56), Fraction(10), Spacing(2, 2, 2, 2))

    styles.margin = 1, 2, 3, 4
    styles.height = "100vh"
    styles.max_height = "50%"

    box_model = widget._get_box_model(Size(60, 20), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(54), Fraction(8), Spacing(1, 2, 3, 4))


def test_max():
    """Check that max_width and max_height are respected."""
    one = Fraction(1)

    class TestWidget(Widget):
        def get_content_width(self, container: Size, parent: Size) -> int:
            assert False, "must not be called"

        def get_content_height(self, container: Size, parent: Size, width: int) -> int:
            assert False, "must not be called"

    widget = TestWidget()
    styles = widget.styles

    styles.width = 100
    styles.height = 80
    styles.max_width = 40
    styles.max_height = 30

    box_model = widget._get_box_model(Size(40, 30), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(40), Fraction(30), Spacing(0, 0, 0, 0))


def test_min():
    """Check that min_width and min_height are respected."""

    one = Fraction(1)

    class TestWidget(Widget):
        def get_content_width(self, container: Size, parent: Size) -> int:
            assert False, "must not be called"

        def get_content_height(self, container: Size, parent: Size, width: int) -> int:
            assert False, "must not be called"

    widget = TestWidget()
    styles = widget.styles
    styles.width = 10
    styles.height = 5
    styles.min_width = 40
    styles.min_height = 30

    box_model = widget._get_box_model(Size(40, 30), Size(80, 24), one, one)
    assert box_model == BoxModel(Fraction(40), Fraction(30), Spacing(0, 0, 0, 0))



from __future__ import annotations

from dataclasses import dataclass
from unittest.mock import Mock

import pytest

from textual._animator import Animator, SimpleAnimation
from textual._easing import DEFAULT_EASING, EASING


class Animatable:
    """An animatable object."""

    def __init__(self, value):
        self.value = value

    def blend(self, destination: Animatable, factor: float) -> Animatable:
        return Animatable(self.value + (destination.value - self.value) * factor)


@dataclass
class AnimateTest:
    """An object with animatable properties."""

    foo: float | None = 0.0  # Plain float that may be set to None on final_value
    bar: Animatable = Animatable(0)  # A mock object supporting the animatable protocol


def test_simple_animation():
    """Test an animation from one float to another."""

    # Thing that may be animated
    animate_test = AnimateTest()

    # Fake wall-clock time
    time = 100.0

    # Object that does the animation
    animation = SimpleAnimation(
        animate_test,
        "foo",
        time,
        3.0,
        start_value=20.0,
        end_value=50.0,
        final_value=None,
        easing=lambda x: x,
    )

    assert animate_test.foo == 0.0

    assert animation(time) is False
    assert animate_test.foo == 20.0

    assert animation(time + 1.0) is False
    assert animate_test.foo == 30.0

    assert animation(time + 2.0) is False
    assert animate_test.foo == 40.0

    assert animation(time + 2.9) is False  # Not quite final value
    assert animate_test.foo == pytest.approx(49.0)

    assert animation(time + 3.0) is True  # True to indicate animation is complete
    assert animate_test.foo is None  # This is final_value

    assert animation(time + 3.0) is True
    assert animate_test.foo is None


def test_simple_animation_duration_zero():
    """Test animation handles duration of 0."""

    # Thing that may be animated
    animatable = AnimateTest()

    # Fake wall-clock time
    time = 100.0

    # Object that does the animation
    animation = SimpleAnimation(
        animatable,
        "foo",
        time,
        0.0,
        start_value=20.0,
        end_value=50.0,
        final_value=50.0,
        easing=lambda x: x,
    )

    assert animation(time) is True  # Duration is 0, so this is last value
    assert animatable.foo == 50.0

    assert animation(time + 1.0) is True
    assert animatable.foo == 50.0


def test_simple_animation_reverse():
    """Test an animation from one float to another, where the end value is less than the start."""

    # Thing that may be animated
    animate_Test = AnimateTest()

    # Fake wall-clock time
    time = 100.0

    # Object that does the animation
    animation = SimpleAnimation(
        animate_Test,
        "foo",
        time,
        3.0,
        start_value=50.0,
        end_value=20.0,
        final_value=20.0,
        easing=lambda x: x,
    )

    assert animation(time) is False
    assert animate_Test.foo == 50.0

    assert animation(time + 1.0) is False
    assert animate_Test.foo == 40.0

    assert animation(time + 2.0) is False
    assert animate_Test.foo == 30.0

    assert animation(time + 3.0) is True
    assert animate_Test.foo == 20.0


def test_animatable():
    """Test SimpleAnimation works with the Animatable protocol"""

    animate_test = AnimateTest()

    # Fake wall-clock time
    time = 100.0

    # Object that does the animation
    animation = SimpleAnimation(
        animate_test,
        "bar",
        time,
        3.0,
        start_value=Animatable(20.0),
        end_value=Animatable(50.0),
        final_value=Animatable(50.0),
        easing=lambda x: x,
    )

    assert animation(time) is False
    assert animate_test.bar.value == 20.0

    assert animation(time + 1.0) is False
    assert animate_test.bar.value == 30.0

    assert animation(time + 2.0) is False
    assert animate_test.bar.value == 40.0

    assert animation(time + 2.9) is False
    assert animate_test.bar.value == pytest.approx(49.0)

    assert animation(time + 3.0) is True  # True to indicate animation is complete
    assert animate_test.bar.value == 50.0


class MockAnimator(Animator):
    """A mock animator."""

    def __init__(self, *args) -> None:
        super().__init__(*args)
        self._time = 0.0
        self._on_animation_frame_called = False

    def on_animation_frame(self):
        self._on_animation_frame_called = True

    def _get_time(self):
        return self._time


async def test_animator():
    target = Mock()
    animator = MockAnimator(target)
    animate_test = AnimateTest()

    # Animate attribute "foo" on animate_test to 100.0 in 10 seconds
    animator.animate(animate_test, "foo", 100.0, duration=10.0)

    expected = SimpleAnimation(
        animate_test,
        "foo",
        0.0,
        duration=10.0,
        start_value=0.0,
        end_value=100.0,
        final_value=100.0,
        easing=EASING[DEFAULT_EASING],
    )
    assert animator._animations[(id(animate_test), "foo")] == expected
    assert not animator._on_animation_frame_called

    animator()
    assert animate_test.foo == 0

    animator._time = 5
    animator()
    assert animate_test.foo == 50

    # New animation in the middle of an existing one
    animator.animate(animate_test, "foo", 200, duration=1)
    assert animate_test.foo == 50

    animator._time = 6
    animator()
    assert animate_test.foo == 200


def test_bound_animator():
    target = Mock()
    animator = MockAnimator(target)
    animate_test = AnimateTest()

    # Bind an animator so it animates attributes on the given object
    bound_animator = animator.bind(animate_test)

    # Animate attribute "foo" on animate_test to 100.0 in 10 seconds
    bound_animator("foo", 100.0, duration=10)

    expected = SimpleAnimation(
        animate_test,
        "foo",
        0,
        duration=10,
        start_value=0,
        end_value=100,
        final_value=100,
        easing=EASING[DEFAULT_EASING],
    )
    assert animator._animations[(id(animate_test), "foo")] == expected


async def test_animator_on_complete_callback_not_fired_before_duration_ends():
    callback = Mock()
    animate_test = AnimateTest()
    animator = MockAnimator(Mock())

    animator.animate(animate_test, "foo", 200, duration=10, on_complete=callback)

    animator._time = 9
    animator()

    assert not callback.called


async def test_animator_on_complete_callback_fired_at_duration():
    callback = Mock()
    animate_test = AnimateTest()
    mock_app = Mock()
    animator = MockAnimator(mock_app)

    animator.animate(animate_test, "foo", 200, duration=10, on_complete=callback)

    animator._time = 10
    animator()

    # Ensure that the callback is scheduled to run after the duration is up.
    mock_app.call_later.assert_called_once_with(callback)


def test_force_stop_animation():
    callback = Mock()
    animate_test = AnimateTest()
    mock_app = Mock()
    animator = MockAnimator(mock_app)

    animator.animate(animate_test, "foo", 200, duration=10, on_complete=callback)

    assert animator.is_being_animated(animate_test, "foo")
    assert animate_test.foo != 200

    animator.force_stop_animation(animate_test, "foo")

    # The animation of the attribute was force cancelled.
    assert not animator.is_being_animated(animate_test, "foo")
    assert animate_test.foo == 200

    # The on_complete callback was scheduled.
    mock_app.call_later.assert_called_once_with(callback)



from __future__ import annotations

try:
    from tree_sitter import Language, Node, Parser, Tree
    from tree_sitter.binding import Query
    from tree_sitter_languages import get_language, get_parser

    TREE_SITTER = True
except ImportError:
    TREE_SITTER = False

from textual.document._document import Document, EditResult, Location, _utf8_encode
from textual.document._languages import BUILTIN_LANGUAGES


class SyntaxAwareDocumentError(Exception):
    """General error raised when SyntaxAwareDocument is used incorrectly."""


class SyntaxAwareDocument(Document):
    """A wrapper around a Document which also maintains a tree-sitter syntax
    tree when the document is edited.

    The primary reason for this split is actually to keep tree-sitter stuff separate,
    since it isn't supported in Python 3.7. By having the tree-sitter code
    isolated in this subclass, it makes it easier to conditionally import. However,
    it does come with other design flaws (e.g. Document is required to have methods
    which only really make sense on SyntaxAwareDocument).

    If you're reading this and Python 3.7 is no longer supported by Textual,
    consider merging this subclass into the `Document` superclass.
    """

    def __init__(
        self,
        text: str,
        language: str | Language,
    ):
        """Construct a SyntaxAwareDocument.

        Args:
            text: The initial text contained in the document.
            language: The language to use. You can pass a string to use a supported
                language, or pass in your own tree-sitter `Language` object.
        """

        if not TREE_SITTER:
            raise RuntimeError("SyntaxAwareDocument unavailable.")

        super().__init__(text)
        self.language: Language | None = None
        """The tree-sitter Language or None if tree-sitter is unavailable."""

        self._parser: Parser | None = None
        """The tree-sitter Parser or None if tree-sitter is unavailable."""

        # If the language is `None`, then avoid doing any parsing related stuff.
        if isinstance(language, str):
            if language not in BUILTIN_LANGUAGES:
                raise SyntaxAwareDocumentError(f"Invalid language {language!r}")
            # If tree-sitter-languages is not installed properly, get_language
            # and get_parser may raise an OSError when unable to load their
            # resources
            try:
                self.language = get_language(language)
                self._parser = get_parser(language)
            except OSError as e:
                raise SyntaxAwareDocumentError(
                    f"Could not find binaries for {language!r}"
                ) from e
        else:
            self.language = language
            self._parser = Parser()
            self._parser.set_language(language)

        self._syntax_tree: Tree = self._parser.parse(self._read_callable)  # type: ignore
        """The tree-sitter Tree (syntax tree) built from the document."""

    @property
    def language_name(self) -> str | None:
        return self.language.name if self.language else None

    def prepare_query(self, query: str) -> Query | None:
        """Prepare a tree-sitter tree query.

        Queries should be prepared once, then reused.

        To execute a query, call `query_syntax_tree`.

        Args:
            query: The string query to prepare.

        Returns:
            The prepared query.
        """
        if not TREE_SITTER:
            raise SyntaxAwareDocumentError(
                "Couldn't prepare query - tree-sitter is not available on this architecture."
            )

        if self.language is None:
            raise SyntaxAwareDocumentError(
                "Couldn't prepare query - no language assigned."
            )

        return self.language.query(query)

    def query_syntax_tree(
        self,
        query: Query,
        start_point: tuple[int, int] | None = None,
        end_point: tuple[int, int] | None = None,
    ) -> list[tuple["Node", str]]:
        """Query the tree-sitter syntax tree.

        The default implementation always returns an empty list.

        To support querying in a subclass, this must be implemented.

        Args:
            query: The tree-sitter Query to perform.
            start_point: The (row, column byte) to start the query at.
            end_point: The (row, column byte) to end the query at.

        Returns:
            A tuple containing the nodes and text captured by the query.
        """

        if not TREE_SITTER:
            raise SyntaxAwareDocumentError(
                "tree-sitter is not available on this architecture."
            )

        captures_kwargs = {}
        if start_point is not None:
            captures_kwargs["start_point"] = start_point
        if end_point is not None:
            captures_kwargs["end_point"] = end_point

        captures = query.captures(self._syntax_tree.root_node, **captures_kwargs)
        return captures

    def replace_range(self, start: Location, end: Location, text: str) -> EditResult:
        """Replace text at the given range.

        Args:
            start: A tuple (row, column) where the edit starts.
            end: A tuple (row, column) where the edit ends.
            text: The text to insert between start and end.

        Returns:
            The new end location after the edit is complete.
        """
        top, bottom = sorted((start, end))

        # An optimisation would be finding the byte offsets as a single operation rather
        # than doing two passes over the document content.
        start_byte = self._location_to_byte_offset(top)
        start_point = self._location_to_point(top)
        old_end_byte = self._location_to_byte_offset(bottom)
        old_end_point = self._location_to_point(bottom)

        replace_result = super().replace_range(start, end, text)

        text_byte_length = len(_utf8_encode(text))
        end_location = replace_result.end_location
        assert self._syntax_tree is not None
        assert self._parser is not None
        self._syntax_tree.edit(
            start_byte=start_byte,
            old_end_byte=old_end_byte,
            new_end_byte=start_byte + text_byte_length,
            start_point=start_point,
            old_end_point=old_end_point,
            new_end_point=self._location_to_point(end_location),
        )
        # Incrementally parse the document.
        self._syntax_tree = self._parser.parse(
            self._read_callable, self._syntax_tree  # type: ignore[arg-type]
        )

        return replace_result

    def get_line(self, line_index: int) -> str:
        """Return the string representing the line, not including new line characters.

        Args:
            line_index: The index of the line.

        Returns:
            The string representing the line.
        """
        line_string = self[line_index]
        return line_string

    def _location_to_byte_offset(self, location: Location) -> int:
        """Given a document coordinate, return the byte offset of that coordinate.
        This method only does work if tree-sitter was imported, otherwise it returns 0.

        Args:
            location: The location to convert.

        Returns:
            An integer byte offset for the given location.
        """
        lines = self._lines
        row, column = location
        lines_above = lines[:row]
        end_of_line_width = len(self.newline)
        bytes_lines_above = sum(
            len(_utf8_encode(line)) + end_of_line_width for line in lines_above
        )
        if row < len(lines):
            bytes_on_left = len(_utf8_encode(lines[row][:column]))
        else:
            bytes_on_left = 0
        byte_offset = bytes_lines_above + bytes_on_left
        return byte_offset

    def _location_to_point(self, location: Location) -> tuple[int, int]:
        """Convert a document location (row_index, column_index) to a tree-sitter
        point (row_index, byte_offset_from_start_of_row). If tree-sitter isn't available
        returns (0, 0).

        Args:
            location: A location (row index, column codepoint offset)

        Returns:
            The point corresponding to that location (row index, column byte offset).
        """
        lines = self._lines
        row, column = location
        if row < len(lines):
            bytes_on_left = len(_utf8_encode(lines[row][:column]))
        else:
            bytes_on_left = 0
        return row, bytes_on_left

    def _read_callable(self, byte_offset: int, point: tuple[int, int]) -> bytes:
        """A callable which informs tree-sitter about the document content.

        This is passed to tree-sitter which will call it frequently to retrieve
        the bytes from the document.

        Args:
            byte_offset: The number of (utf-8) bytes from the start of the document.
            point: A tuple (row index, column *byte* offset). Note that this differs
                from our Location tuple which is (row_index, column codepoint offset).

        Returns:
            All the utf-8 bytes between the byte_offset/point and the end of the current
                line _including_ the line separator character(s). Returns None if the
                offset/point requested by tree-sitter doesn't correspond to a byte.
        """
        row, column = point
        lines = self._lines
        newline = self.newline

        row_out_of_bounds = row >= len(lines)
        if row_out_of_bounds:
            return b""
        else:
            row_text = lines[row]

        encoded_row = _utf8_encode(row_text)
        encoded_row_length = len(encoded_row)

        if column < encoded_row_length:
            return encoded_row[column:] + _utf8_encode(newline)
        elif column == encoded_row_length:
            return _utf8_encode(newline[0])
        elif column == encoded_row_length + 1:
            if newline == "\r\n":
                return b"\n"

        return b""



from textual.app import App, ComposeResult
from textual.widgets import DataTable

ROWS = [
    ("lane", "swimmer", "country", "time"),
    (5, "Chad le Clos", "South Africa", 51.14),
    (4, "Joseph Schooling", "Singapore", 50.39),
    (2, "Michael Phelps", "United States", 51.14),
    (6, "László Cseh", "Hungary", 51.14),
    (3, "Li Zhuhao", "China", 51.26),
    (8, "Mehdy Metella", "France", 51.58),
    (7, "Tom Shields", "United States", 51.73),
    (10, "Darren Burns", "Scotland", 51.84),
    (1, "Aleksandr Sadovnikov", "Russia", 51.84),
]


class TableApp(App):
    def compose(self) -> ComposeResult:
        yield DataTable()

    def on_mount(self) -> None:
        table = self.query_one(DataTable)
        table.fixed_rows = 1
        table.fixed_columns = 1
        table.focus()
        rows = iter(ROWS)
        column_labels = next(rows)
        for column in column_labels:
            table.add_column(column, key=column)
        for index, row in enumerate(rows):
            table.add_row(*row, label=str(index))


app = TableApp()
if __name__ == "__main__":
    app.run()



"""Provides a scrollable text-logging widget."""

from __future__ import annotations

from collections import deque
from typing import TYPE_CHECKING, NamedTuple, Optional, cast

from rich.console import RenderableType
from rich.highlighter import Highlighter, ReprHighlighter
from rich.measure import measure_renderables
from rich.pretty import Pretty
from rich.protocol import is_renderable
from rich.segment import Segment
from rich.text import Text

from textual.cache import LRUCache
from textual.events import Resize
from textual.geometry import Size
from textual.reactive import var
from textual.scroll_view import ScrollView
from textual.strip import Strip

if TYPE_CHECKING:
    from typing_extensions import Self


class DeferredRender(NamedTuple):
    """A renderable which is awaiting rendering.
    This may happen if a `write` occurs before the width is known.

    The arguments are the same as for `RichLog.write`, as this just
    represents a deferred call to that method.
    """

    content: RenderableType | object
    """The content to render."""
    width: int | None = None
    """The width to render or `None` to use optimal width."""
    expand: bool = False
    """Enable expand to widget width, or `False` to use `width`."""
    shrink: bool = True
    """Enable shrinking of content to fit width."""
    scroll_end: bool | None = None
    """Enable automatic scroll to end, or `None` to use `self.auto_scroll`."""


class RichLog(ScrollView, can_focus=True):
    """A widget for logging Rich renderables and text."""

    DEFAULT_CSS = """
    RichLog{
        background: $surface;
        color: $text;
        overflow-y: scroll;
    }
    """

    max_lines: var[int | None] = var[Optional[int]](None)
    min_width: var[int] = var(78)
    wrap: var[bool] = var(False)
    highlight: var[bool] = var(False)
    markup: var[bool] = var(False)
    auto_scroll: var[bool] = var(True)

    def __init__(
        self,
        *,
        max_lines: int | None = None,
        min_width: int = 78,
        wrap: bool = False,
        highlight: bool = False,
        markup: bool = False,
        auto_scroll: bool = True,
        name: str | None = None,
        id: str | None = None,
        classes: str | None = None,
        disabled: bool = False,
    ) -> None:
        """Create a `RichLog` widget.

        Args:
            max_lines: Maximum number of lines in the log or `None` for no maximum.
            min_width: Width to use for calls to `write` with no specified `width`.
            wrap: Enable word wrapping (default is off).
            highlight: Automatically highlight content. By default, the `ReprHighlighter` is used.
                To customize highlighting, set `highlight=True` and then set the `highlighter`
                attribute to an instance of `Highlighter`.
            markup: Apply Rich console markup.
            auto_scroll: Enable automatic scrolling to end.
            name: The name of the text log.
            id: The ID of the text log in the DOM.
            classes: The CSS classes of the text log.
            disabled: Whether the text log is disabled or not.
        """
        super().__init__(name=name, id=id, classes=classes, disabled=disabled)
        self.max_lines = max_lines
        """Maximum number of lines in the log or `None` for no maximum."""
        self._start_line: int = 0
        self.lines: list[Strip] = []
        """The lines currently visible in the log."""
        self._line_cache: LRUCache[tuple[int, int, int, int], Strip]
        self._line_cache = LRUCache(1024)
        self._deferred_renders: deque[DeferredRender] = deque()
        """Queue of deferred renderables to be rendered."""
        self.min_width = min_width
        """Minimum width of renderables."""
        self.wrap = wrap
        """Enable word wrapping."""
        self.highlight = highlight
        """Automatically highlight content."""
        self.markup = markup
        """Apply Rich console markup."""
        self.auto_scroll = auto_scroll
        """Automatically scroll to the end on write."""
        self.highlighter: Highlighter = ReprHighlighter()
        """Rich Highlighter used to highlight content when highlight is True"""

        self._widest_line_width = 0
        """The width of the widest line currently in the log."""

        self._size_known = False
        """Flag which is set to True when the size of the RichLog is known,
        indicating we can proceed with rendering deferred writes."""

    def notify_style_update(self) -> None:
        self._line_cache.clear()

    def on_resize(self, event: Resize) -> None:
        if event.size.width and not self._size_known:
            # This size is known for the first time.
            self._size_known = True
            deferred_renders = self._deferred_renders
            while deferred_renders:
                deferred_render = deferred_renders.popleft()
                self.write(*deferred_render)

    def _make_renderable(self, content: RenderableType | object) -> RenderableType:
        """Make content renderable.

        Args:
            content: Content to render.

        Returns:
            A Rich renderable.
        """
        renderable: RenderableType
        if not is_renderable(content):
            renderable = Pretty(content)
        else:
            if isinstance(content, str):
                if self.markup:
                    renderable = Text.from_markup(content)
                else:
                    renderable = Text(content)
                if self.highlight:
                    renderable = self.highlighter(renderable)
            else:
                renderable = cast(RenderableType, content)

        if isinstance(renderable, Text):
            renderable.expand_tabs()

        return renderable

    def write(
        self,
        content: RenderableType | object,
        width: int | None = None,
        expand: bool = False,
        shrink: bool = True,
        scroll_end: bool | None = None,
    ) -> Self:
        """Write a string or a Rich renderable to the bottom of the log.

        Notes:
            The rendering of content will be deferred until the size of the `RichLog` is known.
            This means if you call `write` in `compose` or `on_mount`, the content will not be
            rendered immediately.

        Args:
            content: Rich renderable (or a string).
            width: Width to render, or `None` to use `RichLog.min_width`.
                If specified, `expand` and `shrink` will be ignored.
            expand: Permit expanding of content to the width of the content region of the RichLog.
                If `width` is specified, then `expand` will be ignored.
            shrink: Permit shrinking of content to fit within the content region of the RichLog.
                If `width` is specified, then `shrink` will be ignored.
            scroll_end: Enable automatic scroll to end, or `None` to use `self.auto_scroll`.

        Returns:
            The `RichLog` instance.
        """
        if not self._size_known:
            # We don't know the size yet, so we'll need to render this later.
            # We defer ALL writes until the size is known, to ensure ordering is preserved.
            if isinstance(content, Text):
                content = content.copy()
            self._deferred_renders.append(
                DeferredRender(content, width, expand, shrink, scroll_end)
            )
            return self

        renderable = self._make_renderable(content)
        auto_scroll = self.auto_scroll if scroll_end is None else scroll_end

        console = self.app.console
        render_options = console.options

        if isinstance(renderable, Text) and not self.wrap:
            render_options = render_options.update(overflow="ignore", no_wrap=True)

        if width is not None:
            # Use the width specified by the caller.
            # We ignore `expand` and `shrink` when a width is specified.
            # This also overrides `min_width` set on the RichLog.
            render_width = width
        else:
            # Compute the width based on available information.
            renderable_width = measure_renderables(
                console, render_options, [renderable]
            ).maximum

            render_width = renderable_width
            scrollable_content_width = self.scrollable_content_region.width

            if expand and renderable_width < scrollable_content_width:
                # Expand the renderable to the width of the scrollable content region.
                render_width = max(renderable_width, scrollable_content_width)

            if shrink and renderable_width > scrollable_content_width:
                # Shrink the renderable down to fit within the scrollable content region.
                render_width = min(renderable_width, scrollable_content_width)

            # The user has not supplied a width, so make sure min_width is respected.
            render_width = max(render_width, self.min_width)

        render_options = render_options.update_width(render_width)

        # Render into (possibly) wrapped lines.
        segments = self.app.console.render(renderable, render_options)
        lines = list(Segment.split_lines(segments))

        if not lines:
            self._widest_line_width = max(render_width, self._widest_line_width)
            self.lines.append(Strip.blank(render_width))
        else:
            strips = Strip.from_lines(lines)
            for strip in strips:
                strip.adjust_cell_length(render_width)
            self.lines.extend(strips)

            if self.max_lines is not None and len(self.lines) > self.max_lines:
                self._start_line += len(self.lines) - self.max_lines
                self.refresh()
                self.lines = self.lines[-self.max_lines :]

            # Compute the width after wrapping and trimming
            # TODO - this is wrong because if we trim a long line, the max width
            #  could decrease, but we don't look at which lines were trimmed here.
            self._widest_line_width = max(
                self._widest_line_width,
                max(sum([segment.cell_length for segment in _line]) for _line in lines),
            )

        # Update the virtual size - the width may have changed after adding
        # the new line(s), and the height will definitely have changed.
        self.virtual_size = Size(self._widest_line_width, len(self.lines))

        if auto_scroll:
            self.scroll_end(animate=False)

        return self

    def clear(self) -> Self:
        """Clear the text log.

        Returns:
            The `RichLog` instance.
        """
        self.lines.clear()
        self._line_cache.clear()
        self._start_line = 0
        self._widest_line_width = 0
        self._deferred_renders.clear()
        self.virtual_size = Size(0, len(self.lines))
        self.refresh()
        return self

    def render_line(self, y: int) -> Strip:
        scroll_x, scroll_y = self.scroll_offset
        line = self._render_line(
            scroll_y + y, scroll_x, self.scrollable_content_region.width
        )
        strip = line.apply_style(self.rich_style)
        return strip

    def _render_line(self, y: int, scroll_x: int, width: int) -> Strip:
        if y >= len(self.lines):
            return Strip.blank(width, self.rich_style)

        key = (y + self._start_line, scroll_x, width, self._widest_line_width)
        if key in self._line_cache:
            return self._line_cache[key]

        line = self.lines[y].crop_extend(scroll_x, scroll_x + width, self.rich_style)

        self._line_cache[key] = line
        return line



import asyncio
import time
import sys

import pytest

from textual.app import App

pytestmark = pytest.mark.skipif(
    sys.platform != "win32", reason="We only need to test this on Windows."
)


def test_win_sleep_timer_is_cancellable():
    """Regression test for https://github.com/Textualize/textual/issues/2711.

    When we exit an app with a "long" timer, everything asyncio-related
    should shutdown quickly. So, we create an app with a timer that triggers
    every SLEEP_FOR seconds and we shut the app down immediately after creating
    it. `asyncio` should be done quickly (i.e., the timer was cancelled) and
    thus the total time this takes should be considerably lesser than the time
    we originally set the timer for.
    """

    SLEEP_FOR = 10

    class WindowsIntervalBugApp(App[None]):
        def on_mount(self) -> None:
            self.set_interval(SLEEP_FOR, lambda: None)

        def key_e(self):
            self.exit()

    async def actual_test():
        async with WindowsIntervalBugApp().run_test() as pilot:
            await pilot.press("e")

    start = time.perf_counter()
    asyncio.run(actual_test())
    end = time.perf_counter()
    assert end - start < 1



from textual.app import App, ComposeResult
from textual.containers import VerticalScroll
from textual.widgets import Label


class LoadingApp(App[None]):
    CSS = """
    VerticalScroll {
        height: 20;
    }
    """

    BINDINGS = [("l", "loading")]

    def compose(self) -> ComposeResult:
        with VerticalScroll():
            yield Label("another big label\n" * 30)  # Ensure there's a scrollbar.

    def action_loading(self):
        self.query_one(VerticalScroll).loading = True


async def test_loading_disables_and_remove_scrollbars():
    app = LoadingApp()
    async with app.run_test() as pilot:
        vs = app.query_one(VerticalScroll)
        # Sanity checks:
        assert not vs._check_disabled()

        await pilot.press("l")
        await pilot.pause()

        assert vs._check_disabled()



from textual.app import App
from textual.widgets import Label


class ContentAlignApp(App):
    CSS_PATH = "content_align.tcss"

    def compose(self):
        yield Label("With [i]content-align[/] you can...", id="box1")
        yield Label("...[b]Easily align content[/]...", id="box2")
        yield Label("...Horizontally [i]and[/] vertically!", id="box3")


if __name__ == "__main__":
    app = ContentAlignApp()
    app.run()



from textual.app import App, ComposeResult
from textual.containers import Container, Vertical
from textual.widgets import Button, Label, Static


async def test_remove_single_widget():
    """It should be possible to the only widget on a screen."""
    async with App().run_test() as pilot:
        widget = Static()
        assert not widget.is_attached
        await pilot.app.mount(widget)
        assert widget.is_attached
        assert len(pilot.app.screen._nodes) == 1
        await pilot.app.query_one(Static).remove()
        assert not widget.is_attached
        assert len(pilot.app.screen._nodes) == 0


async def test_many_remove_all_widgets():
    """It should be possible to remove all widgets on a multi-widget screen."""
    async with App().run_test() as pilot:
        await pilot.app.mount(*[Static() for _ in range(10)])
        assert len(pilot.app.screen._nodes) == 10
        await pilot.app.query(Static).remove()
        assert len(pilot.app.screen._nodes) == 0


async def test_many_remove_some_widgets():
    """It should be possible to remove some widgets on a multi-widget screen."""
    async with App().run_test() as pilot:
        await pilot.app.mount(*[Static(classes=f"is-{n % 2}") for n in range(10)])
        assert len(pilot.app.screen._nodes) == 10
        await pilot.app.query(".is-0").remove()
        assert len(pilot.app.screen._nodes) == 5


async def test_remove_branch():
    """It should be possible to remove a whole branch in the DOM."""
    async with App().run_test() as pilot:
        await pilot.app.mount(
            Container(Container(Container(Container(Container(Static()))))),
            Static(),
            Container(Container(Container(Container(Container(Static()))))),
        )
        assert len(pilot.app.screen.walk_children(with_self=False)) == 13
        await pilot.app.screen._nodes[0].remove()
        assert len(pilot.app.screen.walk_children(with_self=False)) == 7


async def test_remove_overlap():
    """It should be possible to remove an overlapping collection of widgets."""
    async with App().run_test() as pilot:
        await pilot.app.mount(
            Container(Container(Container(Container(Container(Static()))))),
            Static(),
            Container(Container(Container(Container(Container(Static()))))),
        )
        assert len(pilot.app.screen.walk_children(with_self=False)) == 13
        await pilot.app.query(Container).remove()
        assert len(pilot.app.screen.walk_children(with_self=False)) == 1


async def test_remove_move_focus():
    """Removing a focused widget should settle focus elsewhere."""
    async with App().run_test() as pilot:
        buttons = [Button(str(n)) for n in range(10)]
        await pilot.app.mount(Container(*buttons[:5]), Container(*buttons[5:]))
        assert len(pilot.app.screen._nodes) == 2
        assert len(pilot.app.screen.walk_children(with_self=False)) == 12
        assert pilot.app.focused is None
        await pilot.press("tab")
        assert pilot.app.focused is not None
        assert pilot.app.focused == buttons[0]
        await pilot.app.screen._nodes[0].remove()
        assert len(pilot.app.screen._nodes) == 1
        assert len(pilot.app.screen.walk_children(with_self=False)) == 6
        assert pilot.app.focused is not None
        assert pilot.app.focused == buttons[9]


async def test_widget_remove_order() -> None:
    """A Widget.remove of a top-level widget should cause bottom-first removal."""

    removals: list[str] = []

    class Removable(Container):
        def on_unmount(self, _):
            removals.append(self.id if self.id is not None else "unknown")

    async with App().run_test() as pilot:
        await pilot.app.mount(
            Removable(Removable(Removable(id="grandchild"), id="child"), id="parent")
        )
        assert len(pilot.app.screen.walk_children(with_self=False)) == 3
        await pilot.app.screen._nodes[0].remove()
        assert len(pilot.app.screen.walk_children(with_self=False)) == 0
        assert removals == ["grandchild", "child", "parent"]


async def test_query_remove_order():
    """A DOMQuery.remove of a top-level widget should cause bottom-first removal."""

    removals: list[str] = []

    class Removable(Container):
        def on_unmount(self, _):
            removals.append(self.id if self.id is not None else "unknown")

    async with App().run_test() as pilot:
        await pilot.app.mount(
            Removable(Removable(Removable(id="grandchild"), id="child"), id="parent")
        )
        assert len(pilot.app.screen.walk_children(with_self=False)) == 3
        await pilot.app.query(Removable).remove()
        assert len(pilot.app.screen.walk_children(with_self=False)) == 0
        assert removals == ["grandchild", "child", "parent"]


class ExampleApp(App):
    def compose(self) -> ComposeResult:
        yield Button("ABC")
        yield Label("Outside of vertical.")
        with Vertical():
            for index in range(5):
                yield Label(str(index))


async def test_widget_remove_children_container():
    app = ExampleApp()
    async with app.run_test():
        container = app.query_one(Vertical)

        # 6 labels in total, with 5 of them inside the container.
        assert len(app.query(Label)) == 6
        assert len(container.children) == 5

        await container.remove_children()

        # The labels inside the container are gone, and the 1 outside remains.
        assert len(app.query(Label)) == 1
        assert len(container.children) == 0


async def test_widget_remove_children_with_star_selector():
    app = ExampleApp()
    async with app.run_test():
        container = app.query_one(Vertical)

        # 6 labels in total, with 5 of them inside the container.
        assert len(app.query(Label)) == 6
        assert len(container.children) == 5

        await container.remove_children("*")

        # The labels inside the container are gone, and the 1 outside remains.
        assert len(app.query(Label)) == 1
        assert len(container.children) == 0


async def test_widget_remove_children_with_string_selector():
    app = ExampleApp()
    async with app.run_test():
        container = app.query_one(Vertical)

        # 6 labels in total, with 5 of them inside the container.
        assert len(app.query(Label)) == 6
        assert len(container.children) == 5

        await app.screen.remove_children("Label")

        # Only the Screen > Label widget is gone, everything else remains.
        assert len(app.query(Button)) == 1
        assert len(app.query(Vertical)) == 1
        assert len(app.query(Label)) == 5


async def test_widget_remove_children_with_type_selector():
    app = ExampleApp()
    async with app.run_test():
        assert len(app.query(Button)) == 1  # Sanity check.
        await app.screen.remove_children(Button)
        assert len(app.query(Button)) == 0


async def test_widget_remove_children_with_selector_does_not_leak():
    app = ExampleApp()
    async with app.run_test():
        container = app.query_one(Vertical)

        # 6 labels in total, with 5 of them inside the container.
        assert len(app.query(Label)) == 6
        assert len(container.children) == 5

        await container.remove_children("Label")

        # The labels inside the container are gone, and the 1 outside remains.
        assert len(app.query(Label)) == 1
        assert len(container.children) == 0


async def test_widget_remove_children_no_children():
    app = ExampleApp()
    async with app.run_test():
        button = app.query_one(Button)

        count_before = len(app.query("*"))
        await button.remove_children()
        count_after = len(app.query("*"))

        assert len(app.query(Button)) == 1  # The button still remains.
        assert (
            count_before == count_after
        )  # No widgets have been removed, since Button has no children.


async def test_widget_remove_children_no_children_match_selector():
    app = ExampleApp()
    async with app.run_test():
        container = app.query_one(Vertical)
        assert len(container.query("Button")) == 0  # Sanity check.

        count_before = len(app.query("*"))
        container_children_before = list(container.children)
        await container.remove_children("Button")

        assert count_before == len(app.query("*"))
        assert container_children_before == list(container.children)
